from reflectorch.inference.preprocess_exp.preprocess import (
    standard_preprocessing,
    StandardPreprocessing,
)
from reflectorch.inference.preprocess_exp.interpolation import interp_reflectivity
from reflectorch.inference.preprocess_exp.attenuation import apply_attenuation_correction
from reflectorch.inference.preprocess_exp.footprint import apply_footprint_correction


from reflectorch.models.networks.mlp_networks import (
    NetworkWithPriors,
    NetworkWithPriorsConvEmb,
    NetworkWithPriorsFnoEmb,
)
from reflectorch.models.networks.residual_net import ResidualMLP


__all__ = [
    "ResidualMLP",
    "NetworkWithPriors",
    "NetworkWithPriorsConvEmb",
    "NetworkWithPriorsFnoEmb",
]



import torch

import numpy as np

from reflectorch.ml.basic_trainer import (
    TrainerCallback,
    Trainer,
)
from reflectorch.ml.utils import is_divisor

__all__ = [
    'SaveBestModel',
    'LogLosses',
]


class SaveBestModel(TrainerCallback):
    """Callback for periodically saving the best model weights

    Args:
        path (str): path for saving the model weights
        freq (int, optional): frequency in iterations at which the current average loss is evaluated. Defaults to 50.
        average (int, optional): number of recent iterations over which the average loss is computed. Defaults to 10.
    """

    def __init__(self, path: str, freq: int = 50, average: int = 10):
        self.path = path
        self.average = average
        self._best_loss = np.inf
        self.freq = freq

    def end_batch(self, trainer: Trainer, batch_num: int) -> None:
        """checks if the current average loss has improved from the previous save, if true the model is saved

        Args:
            trainer (Trainer): the trainer object
            batch_num (int): the current iteration / batch
        """
        if is_divisor(batch_num, self.freq):

            loss = np.mean(trainer.losses['total_loss'][-self.average:])

            if loss < self._best_loss:
                self._best_loss = loss
                self.save(trainer, batch_num)

    def save(self, trainer: Trainer, batch_num: int):
        """saves a dictionary containing the network weights, the learning rates, the losses and the current \
            best loss with its corresponding iteration to the disk

        Args:
            trainer (Trainer): the trainer object
            batch_num (int): the current iteration / batch
        """
        prev_save = trainer.callback_params.pop('saved_iteration', 0)
        trainer.callback_params['saved_iteration'] = batch_num
        save_dict = {
            'model': trainer.model.state_dict(),
            'lrs': trainer.lrs,
            'losses': trainer.losses,
            'prev_save': prev_save,
            'batch_num': batch_num,
            'best_loss': self._best_loss
        }
        torch.save(save_dict, self.path)


class LogLosses(TrainerCallback):
    """Callback for logging the training losses"""
    def end_batch(self, trainer: Trainer, batch_num: int) -> None:
        """log loss at the current iteration

        Args:
            trainer (Trainer): the trainer object
            batch_num (int): the index of the current iteration / batch
        """
        for loss_name, loss_values in trainer.losses.items():
            try:
                trainer.log(f'train/{loss_name}', loss_values[-1])
            except IndexError:
                continue


# -*- coding: utf-8 -*-

import torch
from torch import Tensor

from reflectorch.data_generation.reflectivity.abeles import abeles_compiled, abeles
from reflectorch.data_generation.reflectivity.memory_eff import abeles_memory_eff
from reflectorch.data_generation.reflectivity.numpy_implementations import (
    kinematical_approximation_np,
    abeles_np,
)
from reflectorch.data_generation.reflectivity.smearing import abeles_constant_smearing
from reflectorch.data_generation.reflectivity.smearing_pointwise import abeles_pointwise_smearing
from reflectorch.data_generation.reflectivity.kinematical import kinematical_approximation


def reflectivity(
        q: Tensor,
        thickness: Tensor,
        roughness: Tensor,
        sld: Tensor,
        dq: Tensor = None,
        gauss_num: int = 51,
        constant_dq: bool = False,
        log: bool = False,
        q_shift: Tensor = 0.0,
        r_scale: Tensor = 1.0,
        background: Tensor = 0.0,
        solvent_vf = None,
        solvent_mode = 'fronting',
        abeles_func = None,
        **abeles_kwargs
):
    """Function which computes the reflectivity curves from thin film parameters. 
    By default it uses the fast implementation of the Abeles matrix formalism.

    Args:
        q (Tensor): tensor of momentum transfer (q) values with shape [batch_size, n_points] or [n_points]
        thickness (Tensor): tensor containing the layer thicknesses (ordered from top to bottom) with shape [batch_size, n_layers]
        roughness (Tensor): tensor containing the interlayer roughnesses (ordered from top to bottom) with shape [batch_size, n_layers + 1]
        sld (Tensor): tensor containing the layer SLDs (real or complex; ordered from top to bottom) with shape 
            [batch_size, n_layers + 1] (excluding ambient SLD which is assumed to be 0)  or [batch_size, n_layers + 2] (including ambient SLD; only for the default ``abeles_func='abeles'``) 
        dq (Tensor, optional): tensor of resolutions used for curve smearing with shape [batch_size, 1].
                            Either dq if ``constant_dq`` is ``True`` or dq/q if ``constant_dq`` is ``False``. Defaults to None.
        gauss_num (int, optional): the number of gaussians for curve smearing. Defaults to 51.
        constant_dq (bool, optional): if ``True`` the smearing is constant (constant dq at each point in the curve) 
                                    otherwise the smearing is linear (constant dq/q at each point in the curve). Defaults to False.
        log (bool, optional): if True the base 10 logarithm of the reflectivity curves is returned. Defaults to False.
        q_shift (float or Tensor, optional): misalignment in q.
        r_scale (float or Tensor, optional): normalization factor (scales reflectivity).
        background (float or Tensor, optional): background intensity.
        abeles_func (Callable, optional): a function implementing the simulation of the reflectivity curves, if different than the default Abeles matrix implementation ('abeles'). Defaults to None.
        abeles_kwargs: Additional arguments specific to the chosen `abeles_func`.
    Returns:
        Tensor: the computed reflectivity curves
    """
    abeles_func = abeles_func or abeles
    q = torch.atleast_2d(q) + q_shift
    q = torch.clamp(q, min=0.0)
    
    if solvent_vf is not None:
        num_layers = thickness.shape[-1]
        if solvent_mode == 'fronting':
            assert sld.shape[-1] == num_layers + 2
            assert solvent_vf.shape[-1] == num_layers
            solvent_sld = sld[..., [0]]
            idx = slice(1, num_layers)
            sld[..., idx] = solvent_vf * solvent_sld + (1.0 - solvent_vf) * sld[..., idx]
        elif solvent_mode == 'backing':
            solvent_sld = sld[..., [-1]]
            idx = slice(1, num_layers) if sld.shape[-1] == num_layers + 2 else slice(0, num_layers)
            sld[..., idx] = solvent_vf * solvent_sld + (1.0 - solvent_vf) * sld[..., idx]
        else:
            raise NotImplementedError

    if dq is None:
        reflectivity_curves = abeles_func(q, thickness, roughness, sld, **abeles_kwargs)
    else:
        if dq.shape[-1] > 1:
            reflectivity_curves = abeles_pointwise_smearing(
                    q=q, dq=dq, thickness=thickness, roughness=roughness, sld=sld, 
                    abeles_func=abeles_func, gauss_num=gauss_num,
                    **abeles_kwargs,
            )
        else:
            reflectivity_curves = abeles_constant_smearing(
                q, thickness, roughness, sld,
                dq=dq, gauss_num=gauss_num, constant_dq=constant_dq, abeles_func=abeles_func,
                **abeles_kwargs,
            )

    if isinstance(r_scale, Tensor):
        r_scale = r_scale.view(-1, *[1] * (reflectivity_curves.dim() - 1))
    if isinstance(background, Tensor):
        background = background.view(-1, *[1] * (reflectivity_curves.dim() - 1))

    reflectivity_curves = reflectivity_curves * r_scale + background

    if log:
        reflectivity_curves = torch.log10(reflectivity_curves)

    return reflectivity_curves



from torch import Tensor


from reflectorch.data_generation import BasicDataset
from reflectorch.data_generation.reflectivity import kinematical_approximation
from reflectorch.data_generation.priors import BasicParams
from reflectorch.ml.basic_trainer import DataLoader


__all__ = [
    "ReflectivityDataLoader",
    "MultilayerDataLoader",
]


class ReflectivityDataLoader(BasicDataset, DataLoader):
    """Dataloader for reflectivity data, combining functionality from the ``BasicDataset`` (basic dataset class for reflectivity) and the ``DataLoader`` (which inherits from ``TrainerCallback``) classes"""
    pass


class MultilayerDataLoader(ReflectivityDataLoader):
    """Dataloader for reflectivity curves simulated using the kinematical approximation"""
    def _sample_from_prior(self, batch_size: int):
        return self.prior_sampler.optimized_sample(batch_size)

    def _calc_curves(self, q_values: Tensor, params: BasicParams):
        return kinematical_approximation(q_values, params.thicknesses, params.roughnesses, params.slds)


from typing import Optional, Tuple, Iterable, Any, Union, Type
from collections import defaultdict

from tqdm import tqdm as standard_tqdm
from tqdm.notebook import tqdm as notebook_tqdm
import numpy as np

import torch
from torch.nn import Module

from reflectorch.ml.loggers import Logger, Loggers

from .utils import is_divisor

__all__ = [
    'Trainer',
    'TrainerCallback',
    'DataLoader',
    'PeriodicTrainerCallback',
]


class Trainer(object):
    """Trainer class
    
    Args:
        model (nn.Module): neural network
        loader (DataLoader): data loader
        lr (float): learning rate
        batch_size (int): batch size
        clip_grad_norm (int, optional): maximum norm for gradient clipping if it is not ``None``. Defaults to None.
        logger (Union[Logger, Tuple[Logger, ...], Loggers], optional): logger. Defaults to None.
        optim_cls (Type[torch.optim.Optimizer], optional): Pytorch optimizer. Defaults to torch.optim.Adam.
        optim_kwargs (dict, optional): optimizer arguments. Defaults to None.
    """

    TOTAL_LOSS_KEY: str = 'total_loss'

    def __init__(self,
                 model: Module,
                 loader: 'DataLoader',
                 lr: float,
                 batch_size: int,
                 clip_grad_norm_max: Optional[int] = None,
                 logger: Union[Logger, Tuple[Logger, ...], Loggers] = None,
                 optim_cls: Type[torch.optim.Optimizer] = torch.optim.Adam,
                 optim_kwargs: dict = None,
                 **kwargs
                 ):
        
        self.model = model
        self.loader = loader
        self.batch_size = batch_size
        self.clip_grad_norm_max = clip_grad_norm_max

        self.optim = self.configure_optimizer(optim_cls, lr=lr, **(optim_kwargs or {}))
        self.lrs = []
        self.losses = defaultdict(list)

        self.logger = _init_logger(logger)
        self.callback_params = {}
        
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.init()

    def init(self):
        pass

    def log(self, name: str, data):
        """log data"""
        self.logger.log(name, data)

    def train(self,
                    num_batches: int,
                    callbacks: Union[Tuple['TrainerCallback', ...], 'TrainerCallback'] = (),
                    disable_tqdm: bool = False,
                    use_notebook_tqdm: bool = False,
                    update_tqdm_freq: int = 1,
                    grad_accumulation_steps: int = 1,
                    ):
        """starts the training process

        Args:
            num_batches (int): total number of training iterations
            callbacks (Union[Tuple['TrainerCallback'], 'TrainerCallback']): the trainer callbacks. Defaults to ().
            disable_tqdm (bool, optional): if ``True``, the progress bar is disabled. Defaults to False.
            use_notebook_tqdm (bool, optional): should be set to ``True`` when used in a Jupyter Notebook. Defaults to False.
            update_tqdm_freq (int, optional): frequency for updating the progress bar. Defaults to 10.
            grad_accumulation_steps (int, optional): number of gradient accumulation steps. Defaults to 1.
        """

        if isinstance(callbacks, TrainerCallback):
            callbacks = (callbacks,)

        callbacks = _StackedTrainerCallbacks(list(callbacks) + [self.loader])

        tqdm_class = notebook_tqdm if use_notebook_tqdm else standard_tqdm
        pbar = tqdm_class(range(num_batches), disable=disable_tqdm)

        callbacks.start_training(self)

        for batch_num in pbar:
            self.model.train()

            self.optim.zero_grad()
            total_loss, avr_loss_dict = 0, defaultdict(list)

            for _ in range(grad_accumulation_steps):

                batch_data = self.get_batch_by_idx(batch_num)
                loss_dict = self.get_loss_dict(batch_data)
                loss = loss_dict['loss'] / grad_accumulation_steps
                total_loss += loss.item()
                _update_loss_dict(avr_loss_dict, loss_dict)

                if not torch.isfinite(loss).item():
                    raise ValueError('Loss is not finite!')

                loss.backward()

            if self.clip_grad_norm_max is not None:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=self.clip_grad_norm_max)

            self.optim.step()

            avr_loss_dict = {k: np.mean(v) for k, v in avr_loss_dict.items()}
            self._update_losses(avr_loss_dict, total_loss)

            if not disable_tqdm:
                self._update_tqdm(pbar, batch_num, update_tqdm_freq)

            break_epoch = callbacks.end_batch(self, batch_num)

            if break_epoch:
                break

        callbacks.end_training(self)

    def _update_tqdm(self, pbar, batch_num: int, update_tqdm_freq: int):
        if is_divisor(batch_num, update_tqdm_freq):
            last_loss = np.mean(self.losses[self.TOTAL_LOSS_KEY][-10:])
            pbar.set_description(f'Loss = {last_loss:.2e}')

            postfix = {}
            for key in self.losses.keys():
                if key != self.TOTAL_LOSS_KEY:
                    last_value = self.losses[key][-1]
                    postfix[key] = f'{last_value:.4f}'

            postfix['lr'] = f'{self.lr():.2e}'

            pbar.set_postfix(postfix)

    def get_batch_by_idx(self, batch_num: int) -> Any:
        raise NotImplementedError

    def get_loss_dict(self, batch_data) -> dict:
        raise NotImplementedError

    def _update_losses(self, loss_dict: dict, loss: float) -> None:
        _update_loss_dict(self.losses, loss_dict)
        self.losses[self.TOTAL_LOSS_KEY].append(loss)
        self.lrs.append(self.lr())

    def configure_optimizer(self, optim_cls, lr: float, **kwargs) -> torch.optim.Optimizer:
        """configure the optimizer based on the optimizer class, the learning rate and the optimizer keyword arguments

        Args:
            optim_cls: the class of the optimizer
            lr (float): the learning rate

        Returns:
            torch.optim.Optimizer:
        """
        optim = optim_cls(self.model.parameters(), lr,  **kwargs)
        return optim

    def lr(self, param_group: int = 0) -> float:
        """get the learning rate"""
        return self.optim.param_groups[param_group]['lr']

    def set_lr(self, lr: float, param_group: int = 0) -> None:
        """set the learning rate"""
        self.optim.param_groups[param_group]['lr'] = lr


class TrainerCallback(object):
    """Base class for trainer callbacks
    """
    def start_training(self, trainer: Trainer) -> None:
        """add functionality the start of training

        Args:
            trainer (Trainer): the trainer object
        """
        pass

    def end_training(self, trainer: Trainer) -> None:
        """add functionality at the end of training

        Args:
            trainer (Trainer): the trainer object
        """
        pass

    def end_batch(self, trainer: Trainer, batch_num: int) -> Union[bool, None]:
        """add functionality at the end of the iteration / batch

        Args:
            trainer (Trainer): the trainer object
            batch_num (int): the index of the current iteration / batch

        Returns:
            Union[bool, None]:
        """
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}()'


class DataLoader(TrainerCallback):
    pass


class PeriodicTrainerCallback(TrainerCallback):
    """Base class for trainer callbacks which perform an action periodically after a number of iterations
        
    Args:
        step (int, optional): Number of iterations after which the action is repeated. Defaults to 1.
        last_epoch (int, optional): the last training  iteration for which the action is performed. Defaults to -1.
    """
    def __init__(self, step: int = 1, last_epoch: int = -1):
        self.step = step
        self.last_epoch = last_epoch

    def end_batch(self, trainer: Trainer, batch_num: int) -> Union[bool, None]:
        """add functionality at the end of the iteration / batch

        Args:
            trainer (Trainer): the trainer object
            batch_num (int): the index of the current iteration / batch

        Returns:
            Union[bool, None]:
        """
        if (
                is_divisor(batch_num, self.step) and
                (self.last_epoch == -1 or batch_num < self.last_epoch)
        ):
            return self._end_batch(trainer, batch_num)

    def _end_batch(self, trainer: Trainer, batch_num: int) -> Union[bool, None]:
        pass


class _StackedTrainerCallbacks(TrainerCallback):
    def __init__(self, callbacks: Iterable[TrainerCallback]):
        self.callbacks = tuple(callbacks)

    def start_training(self, trainer: Trainer) -> None:
        for c in self.callbacks:
            c.start_training(trainer)

    def end_training(self, trainer: Trainer) -> None:
        for c in self.callbacks:
            c.end_training(trainer)

    def end_batch(self, trainer: Trainer, batch_num: int) -> Union[bool, None]:
        break_epoch = False
        for c in self.callbacks:
            break_epoch += bool(c.end_batch(trainer, batch_num))
        return break_epoch

    def __repr__(self):
        callbacks = ", ".join(repr(c) for c in self.callbacks)
        return f'StackedTrainerCallbacks({callbacks})'


def _init_logger(logger: Union[Logger, Tuple[Logger, ...], Loggers] = None):
    if not logger:
        return Logger()
    if isinstance(logger, Logger):
        return logger
    return Loggers(*logger)


def _update_loss_dict(loss_dict: dict, new_values: dict):
    for k, v in new_values.items():
        loss_dict[k].append(v.item())



from reflectorch.models.encoders.conv_encoder import (
    ConvEncoder,
    ConvDecoder,
    ConvAutoencoder,
)
from reflectorch.models.encoders.fno import FnoEncoder, SpectralConv1d
from reflectorch.models.encoders.integral_kernel_embedding import IntegralConvEmbedding
from reflectorch.models.encoders.conv_res_net import ConvResidualNet1D


__all__ = [
    "ConvEncoder",
    "ConvDecoder",
    "ConvAutoencoder",
    "ConvResidualNet1D",
    "FnoEncoder",
    "SpectralConv1d",
    "IntegralConvEmbedding",
]



from pathlib import Path
from typing import Tuple

import torch
import safetensors.torch
import os

from reflectorch import *
from reflectorch.runs.config import load_config

__all__ = [
    "train_from_config",
    "get_trainer_from_config",
    "get_paths_from_config",
    "get_callbacks_from_config",
    "get_trainer_by_name",
    "get_callbacks_by_name",
    "convert_pt_to_safetensors",
]


def init_from_conf(conf, **kwargs):
    """Initializes an object with class type, args and kwargs specified in the configuration

    Args:
        conf (dict): configuration dictionary

    Returns:
        Any: the initialized object
    """
    if not conf:
        return
    cls_name = conf['cls']
    if not cls_name:
        return
    cls = globals().get(cls_name)

    if not cls:
        raise ValueError(f'Unknown class {cls_name}')

    conf_args = conf.get('args', [])
    conf_kwargs = conf.get('kwargs', {})
    conf_kwargs.update(kwargs)
    return cls(*conf_args, **conf_kwargs)


def train_from_config(config: dict):
    """Train a model from a configuration dictionary

    Args:
        config (dict): configuration dictionary

    Returns:
        Trainer: the trainer object
    """

    folder_paths = get_paths_from_config(config, mkdir=True)

    trainer = get_trainer_from_config(config, folder_paths)

    callbacks = get_callbacks_from_config(config, folder_paths)

    trainer.train(
        config['training']['num_iterations'],
        callbacks, disable_tqdm=False,
        update_tqdm_freq=config['training']['update_tqdm_freq'],
        grad_accumulation_steps=config['training'].get('grad_accumulation_steps', 1)
    )

    torch.save({
        'paths': folder_paths,
        'losses': trainer.losses,
        'params': config,
    }, folder_paths['losses'])

    return trainer


def get_paths_from_config(config: dict, mkdir: bool = False):
    """Get the directory paths from a configuration dictionary

    Args:
        config (dict): configuration dictionary
        mkdir (bool, optional): option to create a new directory for the saved model weights and losses.

    Returns:
        dict: dictionary containing the folder paths
    """
    root_dir = Path(config['general']['root_dir'] or ROOT_DIR)
    name = config['general']['name']

    assert root_dir.is_dir()

    saved_models_dir = root_dir / 'saved_models'
    saved_losses_dir = root_dir / 'saved_losses'

    if mkdir:
        saved_models_dir.mkdir(exist_ok=True)
        saved_losses_dir.mkdir(exist_ok=True)

    model_path = str((saved_models_dir / f'model_{name}.pt').absolute())

    losses_path = saved_losses_dir / f'{name}_losses.pt'

    return {
        'name': name,
        'model': model_path,
        'losses': losses_path,
        'root': root_dir,
        'saved_models': saved_models_dir,
        'saved_losses': saved_losses_dir,
    }


def get_callbacks_from_config(config: dict, folder_paths: dict = None) -> Tuple['TrainerCallback', ...]:
    """Initializes the training callbacks from a configuration dictionary
    
    Returns:
        tuple: tuple of callbacks
    """
    callbacks = []

    folder_paths = folder_paths or get_paths_from_config(config)

    train_conf = config['training']
    callback_conf = dict(train_conf['callbacks'])
    save_conf = callback_conf.pop('save_best_model')

    if save_conf['enable']:
        save_model = SaveBestModel(folder_paths['model'], freq=save_conf['freq'])
        callbacks.append(save_model)

    for conf in callback_conf.values():
        callback = init_from_conf(conf)

        if callback:
            callbacks.append(callback)

    if 'logger' in train_conf.keys():
        callbacks.append(LogLosses())

    return tuple(callbacks)


def get_trainer_from_config(config: dict, folder_paths: dict = None):
    """Initializes a trainer from a configuration dictionary

    Args:
        config (dict): the configuration dictionary
        folder_paths (dict, optional): dictionary containing the folder paths

    Returns:
        Trainer: the trainer object
    """
    dset = init_dset(config['dset'])

    folder_paths = folder_paths or get_paths_from_config(config)

    model = init_network(config['model']['network'], folder_paths['saved_models'])

    train_conf = config['training']

    optim_cls = getattr(torch.optim, train_conf['optimizer'])

    logger_conf = train_conf.get('logger', None)
    logger = init_from_conf(logger_conf) if logger_conf and logger_conf.get('cls') else None 

    clip_grad_norm_max = train_conf.get('clip_grad_norm_max', None)

    trainer_cls = globals().get(train_conf['trainer_cls']) if 'trainer_cls' in train_conf else PointEstimatorTrainer
    trainer_kwargs = train_conf.get('trainer_kwargs', {})


    trainer = trainer_cls(
        model, dset, train_conf['lr'], train_conf['batch_size'], clip_grad_norm_max=clip_grad_norm_max,
        logger=logger, optim_cls=optim_cls, 
        **trainer_kwargs
    )

    if train_conf.get('train_with_q_input', False) and getattr(trainer, 'train_with_q_input', None) is not None: #only for back-compatibility with configs in older versions
        trainer.train_with_q_input = True

    return trainer


def get_trainer_by_name(config_name, config_dir=None, model_path=None, load_weights: bool = True, inference_device: str = 'cuda'):
    """Initializes a trainer object based on a configuration file (i.e. the model name) and optionally loads \
        saved weights into the network

    Args:
        config_name (str): name of the configuration file
        config_dir (str): path of the configuration directory
        model_path (str, optional): path to the network weights. The default path is 'saved_models' located in the package directory
        load_weights (bool, optional): if True the saved network weights are loaded into the network. Defaults to True.
        inference_device (str, optional): overwrites the device in the configuration file for the purpose of inference on a different device then the training was performed on. Defaults to 'cuda'.

    Returns:
        Trainer: the trainer object
    """
    config = load_config(config_name, config_dir)
    #config['model']['network']['pretrained_name'] = None

    config['model']['network']['device'] = inference_device
    config['dset']['prior_sampler']['kwargs']['device'] = inference_device
    config['dset']['q_generator']['kwargs']['device'] = inference_device

    trainer = get_trainer_from_config(config)

    num_params = sum(p.numel() for p in trainer.model.parameters())

    print(
        f'Model {config_name} loaded. Number of parameters: {num_params / 10 ** 6:.2f} M',
    )

    if not load_weights:
        return trainer

    if not model_path:
        model_name = f'model_{config_name}.pt'
        model_path = SAVED_MODELS_DIR / model_name

    if str(model_path).endswith('.pt'):
        try:
            state_dict = torch.load(model_path, map_location=inference_device, weights_only=False)
        except Exception as err:
            raise RuntimeError(f'Could not load model from {model_path}') from err

        if 'model' in state_dict:
            trainer.model.load_state_dict(state_dict['model'])
        else:
            trainer.model.load_state_dict(state_dict)

    elif str(model_path).endswith('.safetensors'):
        try:
            load_state_dict_safetensors(model=trainer.model, filename=model_path, device=inference_device)
        except Exception as err:
            raise RuntimeError(f'Could not load model from {model_path}') from err
        
    else:
        raise RuntimeError('Weigths file with unknown extension')

    return trainer
 
def get_callbacks_by_name(config_name, config_dir=None):
    """Initializes the trainer callbacks based on a configuration file

    Args:
        config_name (str): name of the configuration file
        config_dir (str): path of the configuration directory
    """
    config = load_config(config_name, config_dir)
    callbacks = get_callbacks_from_config(config)

    return callbacks


def init_network(config: dict, saved_models_dir: Path):
    """Initializes the network based on the configuration dictionary and optionally loades the weights from a pretrained model"""
    device = config.get('device', 'cuda')
    network = init_from_conf(config).to(device)
    network = load_pretrained(network, config.get('pretrained_name', None), saved_models_dir)
    return network


def load_pretrained(model, model_name: str, saved_models_dir: Path):
    """Loads the saved weights into the network"""
    if not model_name:
        return model
    
    if '.' not in model_name:
        model_name = model_name + '.pt'
    model_path = saved_models_dir / model_name

    if not model_path.is_file():
        model_path = saved_models_dir / f'model_{model_name}'

    if not model_path.is_file():
        raise FileNotFoundError(f'File {str(model_path)} does not exist.')
    
    try:
        pretrained = torch.load(model_path, weights_only=False)
    except Exception as err:
        raise RuntimeError(f'Could not load model from {str(model_path)}') from err
    
    if 'model' in pretrained:
        pretrained = pretrained['model']
    try:
        model.load_state_dict(pretrained)
    except Exception as err:
        raise RuntimeError(f'Could not update state dict from {str(model_path)}') from err

    return model


def init_dset(config: dict):
    """Initializes the dataset / dataloader object"""
    dset_cls = globals().get(config['cls']) if 'cls' in config else ReflectivityDataLoader
    dset_kwargs = config.get('kwargs', {})

    prior_sampler = init_from_conf(config['prior_sampler'])
    intensity_noise = init_from_conf(config['intensity_noise'])
    q_generator = init_from_conf(config['q_generator'])
    curves_scaler = init_from_conf(config['curves_scaler']) if 'curves_scaler' in config else None
    smearing = init_from_conf(config['smearing']) if 'smearing' in config else None
    q_noise = init_from_conf(config['q_noise']) if 'q_noise' in config else None

    dset = dset_cls(
        q_generator=q_generator,
        prior_sampler=prior_sampler,
        intensity_noise=intensity_noise,
        curves_scaler=curves_scaler,
        smearing=smearing,
        q_noise=q_noise,
        **dset_kwargs,
    )

    return dset

def split_complex_tensors(state_dict):
    new_state_dict = {}
    for key, tensor in state_dict.items():
        if tensor.is_complex():
            new_state_dict[f"{key}_real"] = tensor.real.clone()
            new_state_dict[f"{key}_imag"] = tensor.imag.clone()
        else:
            new_state_dict[key] = tensor
    return new_state_dict

def recombine_complex_tensors(state_dict):
    new_state_dict = {}
    keys = list(state_dict.keys())
    visited = set()

    for key in keys:
        if key.endswith('_real') or key.endswith('_imag'):
            base_key = key[:-5]
            new_state_dict[base_key] = torch.complex(state_dict[base_key + '_real'], state_dict[base_key + '_imag'])
            visited.add(base_key + '_real')
            visited.add(base_key + '_imag')
        elif key not in visited:
            new_state_dict[key] = state_dict[key]

    return new_state_dict

def convert_pt_to_safetensors(input_dir):
    """Creates '.safetensors' files for all the model state dictionaries inside '.pt' files in the specified directory.

    Args:
        input_dir (str): directory containing model weights 
    """
    if not os.path.isdir(input_dir):
        raise ValueError(f"Input directory {input_dir} does not exist")

    for file_name in os.listdir(input_dir):
        if file_name.endswith('.pt'):
            pt_file_path = os.path.join(input_dir, file_name)
            safetensors_file_path = os.path.join(input_dir, file_name[:-3] + '.safetensors')

            if os.path.exists(safetensors_file_path):
                print(f"Skipping {pt_file_path}, corresponding .safetensors file already exists.")
                continue

            print(f"Converting {pt_file_path} to .safetensors format.")
            data_pt = torch.load(pt_file_path, weights_only=False)
            model_state_dict = data_pt["model"]
            model_state_dict = split_complex_tensors(model_state_dict) #handle tensors with complex dtype which are not natively supported by safetensors

            safetensors.torch.save_file(tensors=model_state_dict, filename=safetensors_file_path)

def convert_files_to_safetensors(files):
    """
    Converts specified .pt files to .safetensors format.

    Args:
        files (str or list of str): Path(s) to .pt files containing model state dictionaries.
    """
    if isinstance(files, str):
        files = [files]

    for pt_file_path in files:
        if not pt_file_path.endswith('.pt'):
            print(f"Skipping {pt_file_path}: not a .pt file.")
            continue

        if not os.path.exists(pt_file_path):
            print(f"File {pt_file_path} does not exist.")
            continue

        safetensors_file_path = pt_file_path[:-3] + '.safetensors'

        if os.path.exists(safetensors_file_path):
            print(f"Skipping {pt_file_path}: .safetensors version already exists.")
            continue

        print(f"Converting {pt_file_path} to .safetensors format.")
        data_pt = torch.load(pt_file_path, weights_only=False)
        model_state_dict = data_pt["model"]
        model_state_dict = split_complex_tensors(model_state_dict)

        safetensors.torch.save_file(tensors=model_state_dict, filename=safetensors_file_path)

def load_state_dict_safetensors(model, filename, device):
    state_dict = safetensors.torch.load_file(filename=filename, device=device)
    state_dict = recombine_complex_tensors(state_dict)
    model.load_state_dict(state_dict)


from typing import Tuple, Dict, Type, List

import torch
from torch import Tensor

from reflectorch.data_generation.priors.base import PriorSampler
from reflectorch.data_generation.priors.params import AbstractParams
from reflectorch.data_generation.priors.no_constraints import (
    DEFAULT_DEVICE,
    DEFAULT_DTYPE,
)

from reflectorch.data_generation.priors.parametric_models import (
    MULTILAYER_MODELS,
    NuisanceParamsWrapper,
    ParametricModel,
)
from reflectorch.data_generation.priors.scaler_mixin import ScalerMixin


class BasicParams(AbstractParams):
    """Parameter class compatible with different parameterizations of the SLD profile. It stores the parameters as well as their minimum and maximum subprior bounds.

        Args:
            parameters (Tensor): the values of the thin film parameters
            min_bounds (Tensor): the minimum subprior bounds of the parameters
            max_bounds (Tensor): the maximum subprior bounds of the parameters
            max_num_layers (int, optional): the maximum number of layers (for box model parameterizations it is the number of layers). Defaults to None.
            param_model (ParametricModel, optional): the parametric model. Defaults to the box model parameterization with number of layers given by max_num_layers.
        """
    
    __slots__ = (
        'parameters',
        'min_bounds',
        'max_bounds',
        'max_num_layers',
        'param_model',
    )
    PARAM_NAMES = __slots__
    PARAM_MODEL_CLS: Type[ParametricModel]
    MAX_NUM_LAYERS: int = 30

    def __init__(self,
                 parameters: Tensor,
                 min_bounds: Tensor,
                 max_bounds: Tensor,
                 max_num_layers: int = None,
                 param_model: ParametricModel = None,
                 ):

        max_num_layers = max_num_layers or self.MAX_NUM_LAYERS
        self.param_model = param_model or self.PARAM_MODEL_CLS(max_num_layers)
        self.max_num_layers = max_num_layers
        self.parameters = parameters
        self.min_bounds = min_bounds
        self.max_bounds = max_bounds

    def get_param_labels(self, **kwargs) -> List[str]:
        """gets the parameter labels"""
        return self.param_model.get_param_labels(**kwargs)

    def reflectivity(self, q: Tensor, log: bool = False, **kwargs):
        r"""computes the reflectivity curves directly from the parameters

        Args:
            q (Tensor): the q values
            log (bool, optional): whether to apply logarithm to the curves. Defaults to False.

        Returns:
            Tensor: the simulated reflectivity curves
        """
        return self.param_model.reflectivity(q, self.parameters, log=log, **kwargs)

    @property
    def max_layer_num(self) -> int:  # keep for back compatibility but TODO: unify api among different params
        """gets the maximum number of layers"""
        return self.max_num_layers

    @property
    def num_params(self) -> int:
        """get the number of parameters (parameter dimensionality)"""
        return self.param_model.param_dim

    @property
    def thicknesses(self):
        """gets the thicknesses"""
        params = self.param_model.to_standard_params(self.parameters)
        return params['thickness']

    @property
    def roughnesses(self):
        """gets the roughnesses"""
        params = self.param_model.to_standard_params(self.parameters)
        return params['roughness']

    @property
    def slds(self):
        """gets the slds"""
        params = self.param_model.to_standard_params(self.parameters)
        return params['sld']
        
    @property
    def real_slds(self):
        """gets the real part of the slds"""
        params = self.param_model.to_standard_params(self.parameters)
        return params['sld'].real
    
    @property
    def imag_slds(self):
        """gets the imaginary part of the slds (only for complex dtypes)"""
        params = self.param_model.to_standard_params(self.parameters)
        return params['sld'].imag

    @staticmethod
    def rearrange_context_from_params(
            scaled_params: Tensor,
            context: Tensor,
            inference: bool = False,
            from_params: bool = False,
    ):
        if inference:
            if from_params:
                num_params = scaled_params.shape[-1] // 3
                scaled_params = scaled_params[:, num_params:]
            context = torch.cat([context, scaled_params], dim=-1)
            return context

        num_params = scaled_params.shape[-1] // 3
        assert num_params * 3 == scaled_params.shape[-1]
        scaled_params, bound_context = torch.split(scaled_params, [num_params, 2 * num_params], dim=-1)
        context = torch.cat([context, bound_context], dim=-1)
        return scaled_params, context

    @staticmethod
    def restore_params_from_context(scaled_params: Tensor, context: Tensor):
        num_params = scaled_params.shape[-1]
        scaled_bounds = context[:, -2 * num_params:]
        scaled_params = torch.cat([scaled_params, scaled_bounds], dim=-1)
        return scaled_params

    def as_tensor(self, add_bounds: bool = True, **kwargs) -> Tensor:
        """converts the instance of the class to a Pytorch tensor

        Args:
            add_bounds (bool, optional): whether to add the subprior bounds to the tensor. Defaults to True.

        Returns:
            Tensor: the Pytorch tensor obtained from the instance of the class
        """
        if not add_bounds:
            return self.parameters
        return torch.cat([self.parameters, self.min_bounds, self.max_bounds], -1)

    @classmethod
    def from_tensor(cls, params: Tensor, **kwargs):
        """initializes an instance of the class from a Pytorch tensor

        Args:
            params (Tensor): Pytorch tensor containing the parameter values, min subprior bounds and max subprior bounds

        Returns:
            BasicParams: the instance of the class
        """
        num_params = params.shape[-1] // 3

        params, min_bounds, max_bounds = torch.split(
            params, [num_params, num_params, num_params], dim=-1
        )

        return cls(
            params,
            min_bounds,
            max_bounds,
            **kwargs
        )

    def scale_with_q(self, q_ratio: float):
        """scales the parameters based on the q ratio

        Args:
            q_ratio (float): the scaling ratio
        """
        self.parameters = self.param_model.scale_with_q(self.parameters, q_ratio)
        self.min_bounds = self.param_model.scale_with_q(self.min_bounds, q_ratio)
        self.max_bounds = self.param_model.scale_with_q(self.max_bounds, q_ratio)


class SubpriorParametricSampler(PriorSampler, ScalerMixin):
    PARAM_CLS = BasicParams

    def __init__(self,
                 param_ranges: Dict[str, Tuple[float, float]],
                 bound_width_ranges: Dict[str, Tuple[float, float]],
                 model_name: str,
                 device: torch.device = DEFAULT_DEVICE,
                 dtype: torch.dtype = DEFAULT_DTYPE,
                 max_num_layers: int = 50,
                 logdist: bool = False,
                 scale_params_by_ranges = False,
                 scaled_range: Tuple[float, float] = (-1., 1.),
                 **kwargs
                 ):
        """Prior sampler for the parameters of a parametric model and their subprior bounds

        Args:
            param_ranges (Dict[str, Tuple[float, float]]): dictionary containing the name of each type of parameter together with its range
            bound_width_ranges (Dict[str, Tuple[float, float]]): dictionary containing the name of each type of parameter together with the range for sampling the widths of the subprior interval
            model_name (str): the name of the parametric model
            device (torch.device, optional): the Pytorch device. Defaults to DEFAULT_DEVICE.
            dtype (torch.dtype, optional): the Pytorch data type. Defaults to DEFAULT_DTYPE.
            max_num_layers (int, optional): the maximum number of layers (for box model parameterizations it is the number of layers). Defaults to 50.
            logdist (bool, optional): if True the relative widths of the subprior intervals are sampled uniformly on a logarithmic scale instead of uniformly. Defaults to False.
            scale_params_by_ranges (bool, optional): if True the parameters are scaled with respect to their ranges instead of being scaled with respect to their prior bounds. Defaults to False.
            scaled_range (Tuple[float, float], optional): the range for scaling the parameters. Defaults to (-1., 1.)
        """
        self.scaled_range = scaled_range

        self.shift_param_config = kwargs.pop('shift_param_config', {})

        base_model: ParametricModel = MULTILAYER_MODELS[model_name](max_num_layers, logdist=logdist, **kwargs)
        if any(self.shift_param_config.values()):
            self.param_model = NuisanceParamsWrapper(
                base_model=base_model,
                nuisance_params_config=self.shift_param_config,
                **kwargs,
            )
        else:
            self.param_model = base_model

        self.device = device
        self.dtype = dtype
        self.num_layers = max_num_layers

        self.PARAM_CLS.PARAM_MODEL_CLS = MULTILAYER_MODELS[model_name]
        self.PARAM_CLS.MAX_NUM_LAYERS = max_num_layers

        self._param_dim = self.param_model.param_dim
        self.min_bounds, self.max_bounds, self.min_delta, self.max_delta = self.param_model.init_bounds(
            param_ranges, bound_width_ranges, device=device, dtype=dtype
        )
        
        self.param_ranges = param_ranges
        self.bound_width_ranges = bound_width_ranges
        self.model_name = model_name
        self.logdist = logdist
        self.scale_params_by_ranges = scale_params_by_ranges

    @property
    def max_num_layers(self) -> int:
        """gets the maximum number of layers"""
        return self.num_layers

    @property
    def param_dim(self) -> int:
        """get the number of parameters (parameter dimensionality)"""
        return self._param_dim

    def sample(self, batch_size: int) -> BasicParams:
        """sample a batch of parameters

        Args:
            batch_size (int): the batch size

        Returns:
            BasicParams: sampled parameters
        """
        params, min_bounds, max_bounds = self.param_model.sample(
            batch_size, self.min_bounds, self.max_bounds, self.min_delta, self.max_delta
        )

        params = BasicParams(
            parameters=params,
            min_bounds=min_bounds,
            max_bounds=max_bounds,
            max_num_layers=self.max_num_layers,
            param_model=self.param_model,
        )

        return params

    def scale_params(self, params: BasicParams) -> Tensor:
        """scale the parameters to a ML-friendly range

        Args:
            params (BasicParams): the parameters to be scaled

        Returns:
            Tensor: the scaled parameters
        """
        if self.scale_params_by_ranges:
            scaled_params = torch.cat([
                self._scale(params.parameters, self.min_bounds, self.max_bounds), #parameters and subprior bounds are scaled with respect to the parameter ranges
                self._scale(params.min_bounds, self.min_bounds, self.max_bounds), 
                self._scale(params.max_bounds, self.min_bounds, self.max_bounds),
            ], -1)
            return scaled_params
        else:
            scaled_params = torch.cat([
                self._scale(params.parameters, params.min_bounds, params.max_bounds), #each parameter scaled with respect to its subprior bounds
                self._scale(params.min_bounds, self.min_bounds, self.max_bounds), #the subprior bounds are scaled with respect to the parameter ranges
                self._scale(params.max_bounds, self.min_bounds, self.max_bounds),
            ], -1)
            return scaled_params

    def restore_params(self, scaled_params: Tensor) -> BasicParams:
        """restore the parameters to their original range

        Args:
            scaled_params (Tensor): the scaled parameters

        Returns:
            BasicParams: the parameters restored to their original range
        """
        num_params = scaled_params.shape[-1] // 3
        scaled_params, scaled_min_bounds, scaled_max_bounds = torch.split(
            scaled_params, num_params, -1
        )
        if self.scale_params_by_ranges:
            min_bounds = self._restore(scaled_min_bounds, self.min_bounds, self.max_bounds)
            max_bounds = self._restore(scaled_max_bounds, self.min_bounds, self.max_bounds)
            params = self._restore(scaled_params, self.min_bounds, self.max_bounds)
        else:
            min_bounds = self._restore(scaled_min_bounds, self.min_bounds, self.max_bounds)
            max_bounds = self._restore(scaled_max_bounds, self.min_bounds, self.max_bounds)
            params = self._restore(scaled_params, min_bounds, max_bounds)

        return BasicParams(
            parameters=params,
            min_bounds=min_bounds,
            max_bounds=max_bounds,
            max_num_layers=self.max_num_layers,
            param_model=self.param_model,
        )
    
    def scale_bounds(self, bounds: Tensor) -> Tensor:
        return self._scale(bounds, self.min_bounds, self.max_bounds)

    def log_prob(self, params: BasicParams) -> Tensor:
        log_prob = torch.zeros(params.batch_size, device=self.device, dtype=self.dtype)
        log_prob[~self.get_indices_within_bounds(params)] = -float('inf')
        return log_prob

    def get_indices_within_domain(self, params: BasicParams) -> Tensor:
        return self.get_indices_within_bounds(params)

    def get_indices_within_bounds(self, params: BasicParams) -> Tensor:
        return (
                torch.all(params.parameters >= params.min_bounds, -1) &
                torch.all(params.parameters <= params.max_bounds, -1)
        )

    def filter_params(self, params: BasicParams) -> BasicParams:
        indices = self.get_indices_within_domain(params)
        return params[indices]

    def clamp_params(
            self, params: BasicParams, inplace: bool = False
    ) -> BasicParams:
        if inplace:
            params.parameters = torch.clamp_(params.parameters, params.min_bounds, params.max_bounds)
            return params

        return BasicParams(
            parameters=torch.clamp(params.parameters, params.min_bounds, params.max_bounds),
            min_bounds=params.min_bounds.clone(),
            max_bounds=params.max_bounds.clone(),
            max_num_layers=self.max_num_layers,
            param_model=self.param_model,
        )



from reflectorch import *

if __name__ == '__main__':
    config_name = 'a_base_point_xray_conv_standard' #'a_base_point_neutron_conv_standard' 
    trainer = get_trainer_by_name(config_name, load_weights=False)
    trainer.train(1000)


from typing import List, Tuple

import torch
from torch import Tensor

from reflectorch.data_generation.utils import get_d_rhos, get_param_labels
from reflectorch.data_generation.reflectivity import reflectivity

__all__ = [
    "Params",
    "AbstractParams",
]


class AbstractParams(object):
    """Base class for parameters"""
    PARAM_NAMES: Tuple[str, ...]

    @staticmethod
    def rearrange_context_from_params(
            scaled_params: Tensor, context: Tensor, inference: bool = False, from_params: bool = False
    ):
        if inference:
            return context
        return scaled_params, context

    @staticmethod
    def restore_params_from_context(scaled_params: Tensor, context: Tensor):
        return scaled_params

    def reflectivity(self, q: Tensor, log: bool = False, **kwargs):
        """computes the reflectivity curves directly from the parameters"""
        raise NotImplementedError

    def __iter__(self):
        for name in self.PARAM_NAMES:
            yield getattr(self, name)

    def to_(self, tgt):
        for name, arr in zip(self.PARAM_NAMES, self):
            setattr(self, name, _to(arr, tgt))

    def to(self, tgt):
        """performs Pytorch Tensor dtype and/or device conversion"""
        return self.__class__(*[_to(arr, tgt) for arr in self])

    def cuda(self):
        """moves the parameters to the GPU"""
        return self.to('cuda')

    def cpu(self):
        """moves the parameters to the CPU"""
        return self.to('cpu')

    def __getitem__(self, item) -> 'AbstractParams':
        return self.__class__(*[
            arr.__getitem__(item) if isinstance(arr, Tensor) else arr for arr in self
        ])

    def __setitem__(self, key, other):
        if not isinstance(other, AbstractParams):
            raise ValueError

        for param, other_param in zip(self, other):
            if isinstance(param, Tensor):
                param[key] = other_param

    def __add__(self, other):
        if not isinstance(other, AbstractParams):
            raise NotImplemented

        return self.__class__(*[
            torch.cat([param, other_param], 0)
            if isinstance(param, Tensor) else param
            for param, other_param in zip(self, other)
        ])

    def __eq__(self, other):
        if not isinstance(other, AbstractParams):
            raise NotImplemented

        return all([torch.allclose(param, other_param) for param, other_param in zip(self, other)])

    @classmethod
    def cat(cls, *params: 'AbstractParams'):
        return cls(
            *[torch.cat([getattr(p, name) for p in params], 0)
              if isinstance(getattr(params[0], name), Tensor) else
              getattr(params[0], name)
              for name in cls.PARAM_NAMES]
        )

    @property
    def _ref_tensor(self):
        return getattr(self, self.PARAM_NAMES[0])

    def scale_with_q(self, q_ratio: float):
        raise NotImplementedError

    @property
    def max_layer_num(self) -> int:
        """gets the number of layers"""
        return self._ref_tensor.shape[-1]

    @property
    def batch_size(self) -> int:
        """gets the batch size"""
        return self._ref_tensor.shape[0]

    @property
    def device(self):
        """gets the Pytorch device"""
        return self._ref_tensor.device

    @property
    def dtype(self):
        """gets the Pytorch data type"""
        return self._ref_tensor.dtype

    @property
    def d_rhos(self):
        raise NotImplementedError

    def as_tensor(self, use_drho: bool = False) -> Tensor:
        """converts the instance of the class to a Pytorch tensor"""
        raise NotImplementedError

    @classmethod
    def from_tensor(cls, params: Tensor):
        """initializes an instance of the class from a Pytorch tensor"""
        raise NotImplementedError

    @property
    def num_params(self) -> int:
        """get the number of parameters"""
        return self.layers_num2size(self.max_layer_num)

    @staticmethod
    def size2layers_num(size: int) -> int:
        """converts the number of parameters to the number of layers"""
        raise NotImplementedError

    @staticmethod
    def layers_num2size(layers_num: int) -> int:
        """converts the number of layers to the number of parameters"""
        raise NotImplementedError

    def get_param_labels(self) -> List[str]:
        """gets the parameter labels"""
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}(' \
               f'batch_size={self.batch_size}, ' \
               f'max_layer_num={self.max_layer_num}, ' \
               f'device={str(self.device)})'


def _to(arr, dest):
    if hasattr(arr, 'to'):
        arr = arr.to(dest)
    return arr


class Params(AbstractParams):
    """Parameter class for thickness, roughness and sld parameters

        Args:
            thicknesses (Tensor): batch of thicknesses (top to bottom)
            roughnesses (Tensor): batch of roughnesses (top to bottom)
            slds (Tensor): batch of slds (top to bottom)
        """
    MIN_THICKNESS: float = 0.5

    __slots__ = ('thicknesses', 'roughnesses', 'slds')
    PARAM_NAMES = __slots__

    def __init__(self, thicknesses: Tensor, roughnesses: Tensor, slds: Tensor):

        self.thicknesses = thicknesses
        self.roughnesses = roughnesses
        self.slds = slds

    @staticmethod
    def rearrange_context_from_params(
            scaled_params: Tensor, context: Tensor, inference: bool = False, from_params: bool = False
    ):
        if inference:
            return context
        return scaled_params, context

    @staticmethod
    def restore_params_from_context(scaled_params: Tensor, context: Tensor):
        return scaled_params

    def reflectivity(self, q: Tensor, log: bool = False, **kwargs):
        """computes the reflectivity curves directly from the parameters

        Args:
            q (Tensor): the q values
            log (bool, optional): whether to apply logarithm to the curves. Defaults to False.

        Returns:
            Tensor: the simulated reflectivity curves
        """
        return reflectivity(q, self.thicknesses, self.roughnesses, self.slds, log=log, **kwargs)

    def scale_with_q(self, q_ratio: float):
        """scales the parameters based on the q ratio

        Args:
            q_ratio (float): the scaling ratio
        """
        self.thicknesses /= q_ratio
        self.roughnesses /= q_ratio
        self.slds *= q_ratio ** 2

    @property
    def d_rhos(self):
        """computes the differences in SLD values of the neighboring layers"""
        return get_d_rhos(self.slds)

    def as_tensor(self, use_drho: bool = False) -> Tensor:
        """converts the instance of the class to a Pytorch tensor"""
        if use_drho:
            return torch.cat([self.thicknesses, self.roughnesses, self.d_rhos], -1)
        else:
            return torch.cat([self.thicknesses, self.roughnesses, self.slds], -1)

    @classmethod
    def from_tensor(cls, params: Tensor):
        """initializes an instance of the class from a Pytorch tensor containing the values of the parameters"""
        layers_num = (params.shape[-1] - 2) // 3

        thicknesses, roughnesses, slds = torch.split(params, [layers_num, layers_num + 1, layers_num + 1], dim=-1)

        return cls(thicknesses, roughnesses, slds)

    @staticmethod
    def size2layers_num(size: int) -> int:
        """converts the number of parameters to the number of layers"""
        return (size - 2) // 3

    @staticmethod
    def layers_num2size(layers_num: int) -> int:
        """converts the number of layers to the number of parameters"""
        return layers_num * 3 + 2

    def get_param_labels(self, **kwargs) -> List[str]:
        """gets the parameter labels, the layers are numbered from the bottom to the top 
           (i.e. opposite to the order in the Tensors)"""
        return get_param_labels(self.max_layer_num, **kwargs)



import click

from reflectorch.runs.slurm_utils import save_sbatch_and_run
from reflectorch.runs.utils import train_from_config
from reflectorch.runs.config import load_config

__all__ = [
    'run_train',
    'run_train_on_cluster',
    'run_test_config',
]


@click.command()
@click.argument('config_name', type=str)
def run_train(config_name: str):
    """Runs the training from the command line interface
       Example: python -m reflectorch.train 'conf_name'

    Args:
        config_name (str): name of the YAML configuration file
    """
    config = load_config(config_name)
    train_from_config(config)


@click.command()
@click.argument('config_name', type=str)
@click.argument('batch_size', type=int, default=512)
@click.argument('num_iterations', type=int, default=10)
def run_test_config(config_name: str, batch_size: int, num_iterations: int):
    """Run for the purpose of testing the configuration file.
       Example: python -m reflectorch.test_config 'conf_name.yaml' 512 10

    Args:
        config_name (str): name of the YAML configuration file
        batch_size (int): overwrites the batch size in the configuration file
        num_iterations (int): overwrites the number of iterations in the configuration file
    """
    config = load_config(config_name)
    config = _change_to_test_config(config, batch_size=batch_size, num_iterations=num_iterations)
    train_from_config(config)


@click.command()
@click.argument('config_name')
def run_train_on_cluster(config_name: str):
    config = load_config(config_name)
    name = config['general']['name']
    slurm_conf = config['slurm']

    res = save_sbatch_and_run(
        name,
        config_name,
        time=slurm_conf['time'],
        partition=slurm_conf['partition'],
        reservation=slurm_conf.get('reservation', False),
        chdir=slurm_conf.get('chdir', '~/maxwell_output'),
        run_dir=slurm_conf.get('run_dir', None),
        confirm=slurm_conf.get('confirm', True),
    )
    if not res:
        print('Aborted.')
        return
    out, err = res

    if err:
        print('Error occurred: ', err)
    else:
        print('Success!', out)


def _change_to_test_config(config, batch_size: int, num_iterations: int):
    config = dict(config)
    config['training']['num_iterations'] = num_iterations
    config['training']['batch_size'] = batch_size
    config['training']['update_tqdm_freq'] = 1
    return config



# -*- coding: utf-8 -*-
from math import pi, sqrt, log

import torch
from torch import Tensor

from reflectorch.data_generation.reflectivity.abeles import abeles
from torch.nn.functional import conv1d, pad


def abeles_constant_smearing(
        q: Tensor,
        thickness: Tensor,
        roughness: Tensor,
        sld: Tensor,
        dq: Tensor = None,
        gauss_num: int = 31,
        constant_dq: bool = False,
        abeles_func=None,
        **abeles_kwargs
):
    abeles_func = abeles_func or abeles

    if dq.dtype != thickness.dtype:
        q = q.to(thickness)
    
    if dq.dtype != thickness.dtype:
        dq = dq.to(thickness)

    if q.shape[0] == 1:
        q = q.repeat(thickness.shape[0], 1)

    q_lin = _get_q_axes(q, dq, gauss_num, constant_dq=constant_dq)
    kernels = _get_t_gauss_kernels(dq, gauss_num)
    
    curves = abeles_func(q_lin, thickness, roughness, sld, **abeles_kwargs)

    padding = (kernels.shape[-1] - 1) // 2
    padded_curves = pad(curves, (padding, padding), 'reflect')

    smeared_curves = conv1d(
        padded_curves, kernels[:, None], groups=kernels.shape[0],
    )

    if q.shape[0] != smeared_curves.shape[0]:
        repeat_factor = smeared_curves.shape[0] // q.shape[0]
        q = q.repeat(repeat_factor, 1)
        q_lin = q_lin.repeat(repeat_factor, 1)
    
    smeared_curves = _batch_linear_interp1d(q_lin, smeared_curves, q)
        
    return smeared_curves


_FWHM = 2 * sqrt(2 * log(2.0))
_2PI_SQRT = 1. / sqrt(2 * pi)


def _batch_linspace(start: Tensor, end: Tensor, num: int):
    return torch.linspace(0, 1, int(num), device=end.device, dtype=end.dtype)[None] * (end - start) + start


def _torch_gauss(x, s):
    return _2PI_SQRT / s * torch.exp(-0.5 * x ** 2 / s / s)


def _get_t_gauss_kernels(resolutions: Tensor, gaussnum: int = 51):
    gauss_x = _batch_linspace(-1.7 * resolutions, 1.7 * resolutions, gaussnum)
    gauss_y = _torch_gauss(gauss_x, resolutions / _FWHM) * (gauss_x[:, 1] - gauss_x[:, 0])[:, None]
    return gauss_y


def _get_q_axes(q: Tensor, resolutions: Tensor, gaussnum: int = 51, constant_dq: bool = False):
    if constant_dq:
        return _get_q_axes_for_constant_dq(q, resolutions, gaussnum)
    else:
        return _get_q_axes_for_linear_dq(q, resolutions, gaussnum)


def _get_q_axes_for_linear_dq(q: Tensor, resolutions: Tensor, gaussnum: int = 51):
    gaussgpoint = (gaussnum - 1) / 2

    lowq = torch.clamp_min_(q.min(1).values, 1e-6)
    highq = q.max(1).values

    start = torch.log10(lowq)[:, None] - 6 * resolutions / _FWHM
    end = torch.log10(highq[:, None] * (1 + 6 * resolutions / _FWHM))

    interpnums = torch.abs(
        (torch.abs(end - start)) / (1.7 * resolutions / _FWHM / gaussgpoint)
    ).round().to(int)

    q_lin = 10 ** _batch_linspace_with_padding(start, end, interpnums)

    return q_lin


def _get_q_axes_for_constant_dq(q: Tensor, resolutions: Tensor, gaussnum: int = 51) -> Tensor:
    gaussgpoint = (gaussnum - 1) / 2

    start = q.min(1).values[:, None] - resolutions * 1.7
    end = q.max(1).values[:, None] + resolutions * 1.7

    interpnums = torch.abs(
        (torch.abs(end - start)) / (1.7 * resolutions / gaussgpoint)
    ).round().to(int)

    q_lin = _batch_linspace_with_padding(start, end, interpnums)
    q_lin = torch.clamp_min_(q_lin, 1e-6)

    return q_lin


def _batch_linspace_with_padding(start: Tensor, end: Tensor, nums: Tensor) -> Tensor:
    max_num = nums.max().int().item()

    deltas = 1 / (nums - 1)

    x = torch.clamp_min_(_batch_linspace(deltas * (nums - max_num), torch.ones_like(deltas), max_num), 0)

    x = x * (end - start) + start

    return x


def _batch_linear_interp1d(x: Tensor, y: Tensor, x_new: Tensor) -> Tensor:
    eps = torch.finfo(y.dtype).eps

    ind = torch.searchsorted(x.contiguous(), x_new.contiguous())

    ind = torch.clamp_(ind - 1, 0, x.shape[-1] - 2)
    slopes = (y[..., 1:] - y[..., :-1]) / (eps + (x[..., 1:] - x[..., :-1]))
    ind_y = ind + torch.arange(slopes.shape[0], device=slopes.device)[:, None] * y.shape[1]
    ind_slopes = ind + torch.arange(slopes.shape[0], device=slopes.device)[:, None] * slopes.shape[1]

    y_new = y.flatten()[ind_y] + slopes.flatten()[ind_slopes] * (x_new - x.flatten()[ind_y])

    return y_new



import logging
from functools import lru_cache
from typing import Tuple
from math import sqrt

import torch
from torch import Tensor

from reflectorch.data_generation.utils import (
    get_slds_from_d_rhos,
    uniform_sampler,
)

from reflectorch.data_generation.priors.utils import (
    get_allowed_roughness_indices,
    generate_roughnesses,
    params_within_bounds,
)
from reflectorch.data_generation.priors.scaler_mixin import ScalerMixin
from reflectorch.data_generation.priors.base import PriorSampler
from reflectorch.data_generation.priors.params import Params

__all__ = [
    "BasicPriorSampler",
    "DEFAULT_ROUGHNESS_RANGE",
    "DEFAULT_THICKNESS_RANGE",
    "DEFAULT_SLD_RANGE",
    "DEFAULT_NUM_LAYERS",
    "DEFAULT_DEVICE",
    "DEFAULT_DTYPE",
    "DEFAULT_SCALED_RANGE",
    "DEFAULT_USE_DRHO",
]

DEFAULT_THICKNESS_RANGE: Tuple[float, float] = (1., 500.)
DEFAULT_ROUGHNESS_RANGE: Tuple[float, float] = (0., 50.)
DEFAULT_SLD_RANGE: Tuple[float, float] = (-10., 30.)
DEFAULT_NUM_LAYERS: int = 5
DEFAULT_USE_DRHO: bool = False
DEFAULT_DEVICE: torch.device = torch.device('cuda')
DEFAULT_DTYPE: torch.dtype = torch.float64
DEFAULT_SCALED_RANGE: Tuple[float, float] = (-sqrt(3.), sqrt(3.))


class BasicPriorSampler(PriorSampler, ScalerMixin):
    """Prior samplers for thicknesses, roughnesses and slds"""
    def __init__(self,
                 thickness_range: Tuple[float, float] = DEFAULT_THICKNESS_RANGE,
                 roughness_range: Tuple[float, float] = DEFAULT_ROUGHNESS_RANGE,
                 sld_range: Tuple[float, float] = DEFAULT_SLD_RANGE,
                 num_layers: int = DEFAULT_NUM_LAYERS,
                 use_drho: bool = DEFAULT_USE_DRHO,
                 device: torch.device = DEFAULT_DEVICE,
                 dtype: torch.dtype = DEFAULT_DTYPE,
                 scaled_range: Tuple[float, float] = DEFAULT_SCALED_RANGE,
                 restrict_roughnesses: bool = True,
                 ):
        self.logger = logging.getLogger(__name__)
        self.thickness_range = thickness_range
        self.roughness_range = roughness_range
        self.sld_range = sld_range
        self.num_layers = num_layers
        self.device = device
        self.dtype = dtype
        self.scaled_range = scaled_range
        self.use_drho = use_drho
        self.restrict_roughnesses = restrict_roughnesses

    @property
    def max_num_layers(self) -> int:
        return self.num_layers

    @lru_cache()
    def min_vector(self, layers_num, drho: bool = False):
        if drho:
            sld_min = self.sld_range[0] - self.sld_range[1]
        else:
            sld_min = self.sld_range[0]

        return torch.tensor(
            [self.thickness_range[0]] * layers_num +
            [self.roughness_range[0]] * (layers_num + 1) +
            [sld_min] * (layers_num + 1),
            device=self.device,
            dtype=self.dtype
        )

    @lru_cache()
    def max_vector(self, layers_num, drho: bool = False):
        if drho:
            sld_max = self.sld_range[1] - self.sld_range[0]
        else:
            sld_max = self.sld_range[1]
        return torch.tensor(
            [self.thickness_range[1]] * layers_num +
            [self.roughness_range[1]] * (layers_num + 1) +
            [sld_max] * (layers_num + 1),
            device=self.device,
            dtype=self.dtype
        )

    @lru_cache()
    def delta_vector(self, layers_num, drho: bool = False):
        return self._get_delta_vector(self.min_vector(layers_num, drho), self.max_vector(layers_num, drho))

    def restore_params(self, scaled_params: Tensor) -> Params:
        layers_num = self.PARAM_CLS.size2layers_num(scaled_params.shape[-1])

        params_t = self._restore(
            scaled_params,
            self.min_vector(layers_num, drho=self.use_drho).to(scaled_params),
            self.max_vector(layers_num, drho=self.use_drho).to(scaled_params),
        )

        params = self.PARAM_CLS.from_tensor(params_t)

        if self.use_drho:
            params.slds = get_slds_from_d_rhos(params.slds)

        return params

    def scale_params(self, params: Params) -> Tensor:
        layers_num = params.max_layer_num

        return self._scale(
            params.as_tensor(use_drho=self.use_drho),
            self.min_vector(layers_num, drho=self.use_drho).to(params.thicknesses),
            self.max_vector(layers_num, drho=self.use_drho).to(params.thicknesses),
        )

    def get_indices_within_bounds(self, params: Params) -> Tensor:
        layer_num = params.max_layer_num

        return params_within_bounds(
            params.as_tensor(),
            self.min_vector(layer_num),
            self.max_vector(layer_num),
        )

    def get_indices_within_domain(self, params: Params) -> Tensor:
        if self.restrict_roughnesses:
            indices = (
                    self.get_indices_within_bounds(params) &
                    self.get_allowed_roughness_indices(params)
            )
        else:
            indices = self.get_indices_within_bounds(params)
        return indices

    def clamp_params(self, params: Params) -> Params:
        layer_num = params.max_layer_num
        params = params.as_tensor()
        params = torch.clamp(
            params,
            self.min_vector(layer_num),
            self.max_vector(layer_num),
        )
        params = Params.from_tensor(params)
        return params

    @staticmethod
    def get_allowed_roughness_indices(params: Params) -> Tensor:
        return get_allowed_roughness_indices(params.thicknesses, params.roughnesses)

    def log_prob(self, params: Params) -> Tensor:
        # so far we ignore non-uniform distribution of roughnesses and slds.
        log_prob = torch.zeros(params.batch_size, device=params.device, dtype=params.dtype)
        indices = self.get_indices_within_bounds(params)
        log_prob[~indices] = float('-inf')
        return log_prob

    def sample(self, batch_size: int) -> Params:
        slds = self.generate_slds(batch_size)
        thicknesses = self.generate_thicknesses(batch_size)
        roughnesses = self.generate_roughnesses(thicknesses)

        params = Params(thicknesses, roughnesses, slds)

        return params

    def generate_slds(self, batch_size: int):
        return uniform_sampler(
            *self.sld_range, batch_size,
            self.num_layers + 1,
            device=self.device,
            dtype=self.dtype
        )

    def generate_thicknesses(self, batch_size: int):
        return uniform_sampler(
            *self.thickness_range, batch_size,
            self.num_layers,
            device=self.device,
            dtype=self.dtype
        )

    def generate_roughnesses(self, thicknesses: Tensor) -> Tensor:
        if self.restrict_roughnesses:
            return generate_roughnesses(thicknesses, self.roughness_range)
        else:
            return uniform_sampler(
                *self.roughness_range, thicknesses.shape[0],
                self.num_layers + 1,
                device=self.device,
                dtype=self.dtype
            )



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
from matplotlib.lines import Line2D


def print_prediction_results(prediction_dict, param_names=None, width=10, precision=3, header=True):

    if param_names is None:
        param_names = prediction_dict.get("param_names", [])

    pred = np.asarray(prediction_dict.get("predicted_params_array", []), dtype=float)
    pol = prediction_dict.get("polished_params_array", None)
    pol = np.asarray(pol, dtype=float) if pol is not None else None

    name_w = max(14, max((len(str(n)) for n in param_names), default=14))

    num_fmt = f"{{:>{width}.{precision}f}}"

    if header:
        hdr = f"{'Parameter'.ljust(name_w)}  {'Predicted'.rjust(width)}"
        if pol is not None:
            hdr += f"  {'Polished'.rjust(width)}"
        print(hdr)
        print("-" * len(hdr))

    for i, name in enumerate(param_names):
        pred_val = pred[i] if i < pred.size else float("nan")
        row = f"{str(name).ljust(name_w)}  {num_fmt.format(pred_val)}"
        if pol is not None:
            pol_val = pol[i] if i < pol.size else float("nan")
            row += f"  {num_fmt.format(pol_val)}"
        print(row)


def plot_prediction_results(
    prediction_dict: dict,
    q_exp: np.ndarray,
    curve_exp: np.ndarray,
    sigmas_exp: np.ndarray = None,
    logx=False,
):
    q_pred = prediction_dict['q_plot_pred']
    r_pred = prediction_dict['predicted_curve']
    r_pol = prediction_dict.get('polished_curve', None)

    q_pol = None
    if r_pol is not None:
        if len(r_pol) == len(q_pred):
            q_pol = q_pred
        elif len(r_pol) == len(q_exp):
            q_pol = q_exp

    z_sld = prediction_dict.get('predicted_sld_xaxis', None)
    sld_pred_c = prediction_dict.get('predicted_sld_profile', None)
    sld_pol_c  = prediction_dict.get('sld_profile_polished', None)

    plot_sld = (z_sld is not None) and (sld_pred_c is not None or sld_pol_c is not None)

    sld_is_complex = np.iscomplexobj(sld_pred_c)

    sld_pred_label = 'pred. SLD (Re)' if sld_is_complex else 'pred. SLD'
    sld_pol_label  = 'polished SLD (Re)' if sld_is_complex else 'polished SLD'

    fig, axes = plot_reflectivity(
        q_exp=q_exp, r_exp=curve_exp, yerr=sigmas_exp,
        q_pred=q_pred, r_pred=r_pred,
        q_pol=q_pol,  r_pol=r_pol,
        z_sld=z_sld,
        sld_pred=sld_pred_c.real if sld_pred_c is not None else None,
        sld_pol=sld_pol_c.real  if sld_pol_c  is not None else None,
        sld_pred_label=sld_pred_label,
        sld_pol_label=sld_pol_label,
        plot_sld_profile=plot_sld,
        logx=logx,
    )

    if sld_is_complex and plot_sld:
        ax_r, ax_s = axes
        ax_s.plot(z_sld, sld_pred_c.imag, color='darkgreen', lw=2.0, ls='-', zorder=4, label='pred. SLD (Im)')
        if sld_pol_c is not None:
            ax_s.plot(z_sld, sld_pol_c.imag, color='cyan', lw=2.0, ls='--', zorder=5, label='polished SLD (Im)')
        ax_s.legend(fontsize=14, frameon=True)

    return fig, axes


def plot_reflectivity(
    *,
    q_exp=None, 
    r_exp=None, 
    yerr=None, 
    xerr=None,
    q_pred=None, 
    r_pred=None,
    q_pol=None,  
    r_pol=None,
    z_sld=None, 
    sld_pred=None, 
    sld_pol=None,
    plot_sld_profile=False,
    figsize=None,
    logx=False,
    logy=True,
    x_ticks_log=None,
    y_ticks_log=(10.0 ** -np.arange(0, 12, 2)),
    q_label=r'q [$\mathrm{\AA^{-1}}$]',
    r_label='R(q)',
    z_label=r'z [$\mathrm{\AA}$]',
    sld_label=r'SLD [$10^{-6}\ \mathrm{\AA^{-2}}$]',
    xlim=None,
    axis_label_size=20,
    tick_label_size=15,
    legend_fontsize=14,
    exp_style='auto',
    exp_color='blue',
    exp_facecolor='none',
    exp_marker='o',
    exp_ms=3,
    exp_alpha=1.0,
    exp_errcolor='purple',
    exp_elinewidth=1.0,
    exp_capsize=1.0,
    exp_capthick=1.0,
    exp_zorder=2,
    pred_color='red',
    pred_lw=2.0,
    pred_ls='-',
    pred_alpha=1.0,
    pred_zorder=3,
    pol_color='orange',
    pol_lw=2.0,
    pol_ls='--',
    pol_alpha=1.0,
    pol_zorder=4,
    sld_pred_color='red',   
    sld_pred_lw=2.0,   
    sld_pred_ls='-',
    sld_pol_color='orange', 
    sld_pol_lw=2.0,    
    sld_pol_ls='--',
    exp_label='exp. data',
    pred_label='prediction',
    pol_label='polished prediction',
    sld_pred_label='pred. SLD',
    sld_pol_label='polished SLD',
    legend=True,
    legend_kwargs=None
):

    def _np(a):
        return None if a is None else np.asarray(a)

    def _mask(x, y):
        m = np.isfinite(x) & np.isfinite(y)
        if logx: m &= (x > 0.0)
        if logy: m &= (y > 0.0)
        return m

    def _slice_sym_err(err, mask):
        if err is None:
            return None
        if np.isscalar(err):
            return err
        e = np.asarray(err)
        if e.ndim != 1:
            raise ValueError("Errors must be scalar or 1-D array.")
        return e[mask]

    q_exp, r_exp, yerr, xerr = _np(q_exp), _np(r_exp), _np(yerr), _np(xerr)
    q_pred, r_pred = _np(q_pred), _np(r_pred)
    q_pol,  r_pol  = _np(q_pol),  _np(r_pol)
    z_sld, sld_pred, sld_pol = _np(z_sld), _np(sld_pred), _np(sld_pol)

    # Figure & axes
    if figsize is None:
        figsize = (12, 6) if plot_sld_profile else (6, 6)
    if plot_sld_profile:
        fig, (ax_r, ax_s) = plt.subplots(1, 2, figsize=figsize)
    else:
        fig, ax_r = plt.subplots(1, 1, figsize=figsize)
        ax_s = None

    # Apply x-limits (right-only or both)
    if xlim is not None:
        if np.isscalar(xlim):
            cur_left, _ = ax_r.get_xlim()
            if logx and cur_left <= 0:
                cur_left = 1e-12
            ax_r.set_xlim(left=cur_left, right=float(xlim))
        else:
            xmin, xmax = xlim
            if logx and xmin is not None and xmin <= 0:
                raise ValueError("For log-x, xmin must be > 0.")
            ax_r.set_xlim(left=xmin, right=xmax)

    # Axis scales / labels / ticks
    if logx: ax_r.set_xscale('log')
    if logy: ax_r.set_yscale('log')

    ax_r.set_xlabel(q_label, fontsize=axis_label_size)
    ax_r.set_ylabel(r_label, fontsize=axis_label_size)
    ax_r.tick_params(axis='both', which='major', labelsize=tick_label_size)
    ax_r.tick_params(axis='both', which='minor', labelsize=tick_label_size)
    if logx and x_ticks_log is not None:
        ax_r.xaxis.set_major_locator(mticker.FixedLocator(x_ticks_log))
    if logy and y_ticks_log is not None:
        ax_r.yaxis.set_major_locator(mticker.FixedLocator(y_ticks_log))

    handles = []

    # Experimental plot
    exp_handle = None
    if q_exp is not None and r_exp is not None:
        m = _mask(q_exp, r_exp)
        style = exp_style if exp_style != 'auto' else ('errorbar' if yerr is not None else 'scatter')

        if style == 'errorbar' and (yerr is not None):
            yerr_m = _slice_sym_err(yerr, m)
            xerr_m = _slice_sym_err(xerr, m)
            ax_r.errorbar(
                q_exp[m], r_exp[m], yerr=yerr_m, xerr=xerr_m,
                color=exp_color, ecolor=exp_errcolor,
                elinewidth=exp_elinewidth, capsize=exp_capsize,
                capthick=(exp_elinewidth if exp_capthick is None else exp_capthick),
                marker=exp_marker, linestyle='none', markersize=exp_ms,
                markerfacecolor=exp_facecolor, markeredgecolor=exp_color,
                alpha=exp_alpha, zorder=exp_zorder, label=None
            )
            exp_handle = Line2D([], [], color=exp_color, marker=exp_marker,
                                linestyle='none', markersize=exp_ms,
                                markerfacecolor=exp_facecolor, markeredgecolor=exp_color,
                                alpha=exp_alpha, label=exp_label)
        elif style == 'scatter':
            ax_r.scatter(
                q_exp[m], r_exp[m],
                s=exp_ms**2, marker=exp_marker,
                facecolors=exp_facecolor, edgecolors=exp_color,
                alpha=exp_alpha, zorder=exp_zorder, label=None
            )
            exp_handle = Line2D([], [], color=exp_color, marker=exp_marker,
                                linestyle='none', markersize=exp_ms,
                                markerfacecolor=exp_facecolor, markeredgecolor=exp_color,
                                alpha=exp_alpha, label=exp_label)
        else:  # 'line'
            ln = ax_r.plot(
                q_exp[m], r_exp[m], color=exp_color, lw=1.0, ls='-',
                alpha=exp_alpha, zorder=exp_zorder, label=exp_label
            )[0]
            exp_handle = ln

        if exp_handle is not None:
            handles.append(exp_handle)

    # Predicted line
    pred_handle = None
    if q_pred is not None and r_pred is not None:
        m = _mask(q_pred, r_pred)
        pred_handle = ax_r.plot(
            q_pred[m], r_pred[m],
            color=pred_color, lw=pred_lw, ls=pred_ls,
            alpha=pred_alpha, zorder=pred_zorder, label=pred_label
        )[0]
        handles.append(pred_handle)

    # Polished line
    pol_handle = None
    if q_pol is not None and r_pol is not None:
        m = _mask(q_pol, r_pol)
        pol_handle = ax_r.plot(
            q_pol[m], r_pol[m],
            color=pol_color, lw=pol_lw, ls=pol_ls,
            alpha=pol_alpha, zorder=pol_zorder, label=pol_label
        )[0]
        handles.append(pol_handle)

    if legend and handles:
        lk = {} if legend_kwargs is None else dict(legend_kwargs)
        ax_r.legend(handles=handles,
                    labels=[h.get_label() for h in handles],
                    fontsize=legend_fontsize, loc='best', **lk)

    # SLD panel (optional)
    if ax_s is not None:
        ax_s.set_xlabel(z_label, fontsize=axis_label_size)
        ax_s.set_ylabel(sld_label, fontsize=axis_label_size)
        ax_s.tick_params(axis='both', which='major', labelsize=tick_label_size)
        ax_s.tick_params(axis='both', which='minor', labelsize=tick_label_size)

        if z_sld is not None and sld_pred is not None:
            ax_s.plot(z_sld, sld_pred,
                      color=sld_pred_color, lw=sld_pred_lw, ls=sld_pred_ls,
                      label=sld_pred_label)
        if z_sld is not None and sld_pol is not None:
            ax_s.plot(z_sld, sld_pol,
                      color=sld_pol_color, lw=sld_pol_lw, ls=sld_pol_ls,
                      label=sld_pol_label)

        if legend:
            ax_s.legend(fontsize=legend_fontsize, loc='best', **(legend_kwargs or {}))

    plt.tight_layout()
    return (fig, (ax_r, ax_s)) if ax_s is not None else (fig, ax_r)


def plot_reflectivity_multi(
    *,
    rq_series,
    sld_series=None,
    plot_sld_profile=False,
    figsize=None,
    logx=False,
    logy=True,
    xlim=None,
    x_ticks_log=None,
    y_ticks_log=(10.0 ** -np.arange(0, 12, 2)),
    q_label=r'q [$\mathrm{\AA^{-1}}$]',
    r_label='R(q)',
    z_label=r'z [$\mathrm{\AA}$]',
    sld_label=r'SLD [$10^{-6}\ \mathrm{\AA^{-2}}$]',
    axis_label_size=20,
    tick_label_size=15,
    legend=True,
    legend_fontsize=12,
    legend_kwargs=None,
):
    """
    Plot multiple R(q) series (and optional SLD lines) with per-series styling.

    rq_series: list of dicts, each with:
        required:
            - x: 1D array
            - y: 1D array
        optional (per series):
            - kind: 'errorbar' | 'scatter' | 'line'  (default 'line')
            - label: str
            - color: str
            - alpha: float (0..1)
            - zorder: int
            # scatter / marker for errorbar:
            - marker: str (default 'o')
            - ms: float (marker size in pt; for scatter internally converted to s=ms**2)
            - facecolor: str (scatter marker face)
            # errorbar only:
            - yerr: scalar or 1D array
            - xerr: scalar or 1D array
            - ecolor: str
            - elinewidth: float
            - capsize: float
            - capthick: float
            # line only:
            - lw: float
            - ls: str (e.g. '-', '--', ':')

    sld_series: list of dicts (only lines), each with:
        - x: 1D array (z-axis)
        - y: 1D array (SLD)
        - label: str (optional)
        - color: str (optional)
        - lw: float (optional)
        - ls: str (optional)
        - alpha: float (optional)
        - zorder: int (optional)
    """

    def _np(a): return None if a is None else np.asarray(a)

    def _mask(x, y):
        m = np.isfinite(x) & np.isfinite(y)
        if logx: m &= (x > 0.0)
        if logy: m &= (y > 0.0)
        return m

    # Figure & axes
    if figsize is None:
        figsize = (12, 6) if plot_sld_profile else (6, 6)
    if plot_sld_profile:
        fig, (ax_r, ax_s) = plt.subplots(1, 2, figsize=figsize)
    else:
        fig, ax_r = plt.subplots(1, 1, figsize=figsize)
        ax_s = None

    # Axis scales / labels / ticks
    if logx: ax_r.set_xscale('log')
    if logy: ax_r.set_yscale('log')

    ax_r.set_xlabel(q_label, fontsize=axis_label_size)
    ax_r.set_ylabel(r_label, fontsize=axis_label_size)
    ax_r.tick_params(axis='both', which='major', labelsize=tick_label_size)
    ax_r.tick_params(axis='both', which='minor', labelsize=tick_label_size)
    if logx and x_ticks_log is not None:
        ax_r.xaxis.set_major_locator(mticker.FixedLocator(x_ticks_log))
    if logy and y_ticks_log is not None:
        ax_r.yaxis.set_major_locator(mticker.FixedLocator(y_ticks_log))

    # Apply x-limits (right-only or both)
    if xlim is not None:
        if np.isscalar(xlim):
            left, _ = ax_r.get_xlim()
            if logx and left <= 0:
                left = 1e-12
            ax_r.set_xlim(left=left, right=float(xlim))
        else:
            xmin, xmax = xlim
            if logx and xmin is not None and xmin <= 0:
                raise ValueError("For log-x, xmin must be > 0.")
            ax_r.set_xlim(left=xmin, right=xmax)

    # Plot all R(q) series in the given order (order = legend order)
    handles = []
    for s in rq_series:
        kind  = s.get('kind', 'line')
        x     = _np(s.get('x'))
        y     = _np(s.get('y'))
        if x is None or y is None:
            continue

        label = s.get('label', None)
        color = s.get('color', None)
        alpha = s.get('alpha', 1.0)
        zord  = s.get('zorder', None)
        ms    = s.get('ms', 5.0)
        marker = s.get('marker', 'o')

        m = _mask(x, y)

        if kind == 'errorbar':
            yerr = s.get('yerr', None)
            xerr = s.get('xerr', None)
            ecolor     = s.get('ecolor', color)
            elinewidth = s.get('elinewidth', 1.0)
            capsize    = s.get('capsize', 0.0)
            capthick   = s.get('capthick', elinewidth)

            # Symmetric error input: scalar or 1D
            def _slice_sym(err):
                if err is None: return None
                if np.isscalar(err): return err
                arr = np.asarray(err)
                if arr.ndim != 1:
                    raise ValueError("For symmetric error bars, provide scalar or 1-D array.")
                return arr[m]

            yerr_m = _slice_sym(yerr)
            xerr_m = _slice_sym(xerr)

            ax_r.errorbar(
                x[m], y[m], yerr=yerr_m, xerr=xerr_m,
                color=color, ecolor=ecolor,
                elinewidth=elinewidth, capsize=capsize, capthick=capthick,
                marker=marker, linestyle='none', markersize=ms,
                markerfacecolor=s.get('facecolor', 'none'),
                markeredgecolor=color,
                alpha=alpha, zorder=zord, label=None
            )

            h = Line2D([], [], color=color, marker=marker, linestyle='none',
                       markersize=ms, markerfacecolor=s.get('facecolor','none'),
                       markeredgecolor=color, alpha=alpha, label=label)
            if label is not None:
                handles.append(h)

        elif kind == 'scatter':
            facecolor = s.get('facecolor', 'none')
            sc = ax_r.scatter(
                x[m], y[m], s=ms**2, marker=marker,
                facecolors=facecolor, edgecolors=color,
                alpha=alpha, zorder=zord, label=None
            )
            h = Line2D([], [], color=color, marker=marker, linestyle='none',
                       markersize=ms, markerfacecolor=facecolor,
                       markeredgecolor=color, alpha=alpha, label=label)
            if label is not None:
                handles.append(h)

        else:  # 'line'
            lw = s.get('lw', 2.0)
            ls = s.get('ls', '-')
            line = ax_r.plot(
                x[m], y[m],
                color=color, lw=lw, ls=ls,
                alpha=alpha, zorder=zord, label=label
            )[0]
            if label is not None:
                handles.append(line)

    if legend and handles:
        lk = {} if legend_kwargs is None else dict(legend_kwargs)
        ax_r.legend(handles=handles,
                    labels=[h.get_label() for h in handles],
                    fontsize=legend_fontsize, loc='best', **lk)

    # Optional SLD panel
    if plot_sld_profile:
        ax_s.set_xlabel(z_label, fontsize=axis_label_size)
        ax_s.set_ylabel(sld_label, fontsize=axis_label_size)
        ax_s.tick_params(axis='both', which='major', labelsize=tick_label_size)
        ax_s.tick_params(axis='both', which='minor', labelsize=tick_label_size)

        if sld_series:
            for s in sld_series:
                zx = _np(s.get('x')); zy = _np(s.get('y'))
                if zx is None or zy is None:
                    continue
                label = s.get('label', None)
                color = s.get('color', None)
                lw    = s.get('lw', 2.0)
                ls    = s.get('ls', '-')
                alpha = s.get('alpha', 1.0)
                zord  = s.get('zorder', None)
                ax_s.plot(zx, zy, color=color, lw=lw, ls=ls, alpha=alpha, zorder=zord, label=label)

            if legend:
                ax_s.legend(fontsize=legend_fontsize, loc='best', **(legend_kwargs or {}))

    plt.tight_layout()
    return (fig, (ax_r, ax_s)) if plot_sld_profile else (fig, ax_r)


from reflectorch.ml.basic_trainer import *
from reflectorch.ml.callbacks import *
from reflectorch.ml.trainers import *
from reflectorch.ml.loggers import *
from reflectorch.ml.schedulers import *
from reflectorch.ml.dataloaders import *

__all__ = [
    'Trainer',
    'TrainerCallback',
    'DataLoader',
    'PeriodicTrainerCallback',
    'SaveBestModel',
    'LogLosses',
    'Logger',
    'Loggers',
    'PrintLogger',
    'TensorBoardLogger',
    'ScheduleBatchSize',
    'ScheduleLR',
    'StepLR',
    'CyclicLR',
    'LogCyclicLR',
    'ReduceLROnPlateau',
    'OneCycleLR',
    'CosineAnnealingWithWarmup',
    'ReflectivityDataLoader',
    'MultilayerDataLoader',
    'RealTimeSimTrainer',
    'DenoisingAETrainer',
    'PointEstimatorTrainer',
]



import yaml

from pathlib import Path
from reflectorch.paths import CONFIG_DIR


def load_config(config_name: str, config_dir: str = None) -> dict:
    """Loads a configuration dictionary from a YAML configuration file located in the configuration directory

    Args:
        config_name (str): name of the YAML configuration file
        config_dir (str): path of the configuration directory

    Returns:
        dict: the configuration dictionary
    """
    if not config_name.endswith('.yaml'):
        config_name = f'{config_name}.yaml'
    config_dir = Path(config_dir) if config_dir else CONFIG_DIR
    path = config_dir / config_name
    with open(path, 'r') as f:
        config = yaml.safe_load(f)
    config['config_path'] = str(path.absolute())

    return config



import pytest
import numpy as np

from reflectorch.paths import TEST_DATA_PATH

@pytest.fixture(
    params=["demo_scan.csv"],
    scope="session"
)
def raw_data_with_preprocessing_params(request):
    scan_path = TEST_DATA_PATH / request.param
    data = np.loadtxt(str(scan_path), skiprows=1, delimiter=',')

    raw_data = dict(
        intensity=data[:, 5],
        scattering_angle=data[:, 0] * 2,
        attenuation=data[:, 4],
    )

    preprocessing_params = dict(
        q_interp=np.linspace(0.02, 0.15, 64),
        wavelength=1,
        beam_width=0.2,
        sample_length=10.,
    )

    return raw_data, preprocessing_params


import pytest
import torch
from reflectorch.data_generation.reflectivity import reflectivity, abeles, abeles_memory_eff, kinematical_approximation

@pytest.mark.parametrize("abeles_func", [abeles, abeles_memory_eff, kinematical_approximation])
def test_reflectivity_two_layers(abeles_func):

    thickness_tensor = torch.tensor([
        [50, 100], # element 1 in the batch (n_layers thicknesses from top/ambient to bottom/substrate)
        [80, 40], # element 2 in the batch
        [600, 100], # element 3 in the batch
    ])
    
    roughness_tensor = torch.tensor([
        [1.0, 5.0, 3.0], # (n_layers + 1 roughnesses from top/ambient to bottom/substrate)
        [15.5, 1.0, 2.0],
        [0.0, 0.0, 50.0],
    ])

    sld_tensor = torch.tensor([
        [20.0, 12.1, 18.5], # (n_layers + 1 slds from top/ambient to bottom/substrate)
        [-5.0, 10.5, 12.0],
        [15.0 + 1j*5, 5.99 + 1j*0.09, 25.0],
    ])

    q_min = 0.02
    q_max = 0.15
    n_q = 128 
    q = torch.linspace(q_min, q_max, n_q)

    refl_curves = reflectivity(
        q=q, 
        thickness=thickness_tensor, 
        roughness=roughness_tensor, 
        sld=sld_tensor,
        abeles_func=abeles_func,
    )
    
    batch_size = thickness_tensor.shape[0]

    assert refl_curves.shape == (batch_size, n_q)


@pytest.mark.parametrize("abeles_func", [abeles, abeles_memory_eff, kinematical_approximation])
def test_reflectivity_four_layers(abeles_func):

    thickness_tensor = torch.tensor([
        [50, 100, 10, 100],
        [80, 40, 1, 60],
        [600, 100, 50, 50],
    ])
    
    roughness_tensor = torch.tensor([
        [1.0, 5.0, 3.0, 0.5, 4.0],
        [15.5, 1.0, 2.0, 0.5, 4.0],
        [0.0, 0.0, 50.0, 12.3, 6.2],
    ])

    sld_tensor = torch.tensor([
        [20.0, 12.1, 18.5, 1.0, 25.0],
        [-5.0, 10.5, 12.0, -2.15, 15.0],
        [15.0 + 1j*5, 5.99 + 1j*0.09, 25.0, 10.1 + 1j*4.3, 16.0],
    ])

    q_min = 0.01
    q_max = 0.3
    n_q = 200
    q = torch.linspace(q_min, q_max, n_q)

    refl_curves = reflectivity(
        q=q, 
        thickness=thickness_tensor, 
        roughness=roughness_tensor, 
        sld=sld_tensor,
        abeles_func=abeles_func,
    )
    
    batch_size = thickness_tensor.shape[0]

    assert refl_curves.shape == (batch_size, n_q)

def test_abeles_vs_abelesmemoryeff():

    thickness_tensor = torch.tensor([
        [50, 100, 10, 100],
        [80, 40, 1, 60],
        [600, 100, 50, 50],
    ])
    
    roughness_tensor = torch.tensor([
        [1.0, 5.0, 3.0, 0.5, 4.0],
        [15.5, 1.0, 2.0, 0.5, 4.0],
        [0.0, 0.0, 50.0, 12.3, 6.2],
    ])

    sld_tensor = torch.tensor([
        [20.0, 12.1, 18.5, 1.0, 25.0],
        [-5.0, 10.5, 12.0, -2.15, 15.0],
        [15.0 + 1j*5, 5.99 + 1j*0.09, 25.0, 10.1 + 1j*4.3, 16.0],
    ])

    q_min = 0.01
    q_max = 0.3
    n_q = 200
    q = torch.linspace(q_min, q_max, n_q)

    refl_curves_abeles = reflectivity(
        q=q, 
        thickness=thickness_tensor, 
        roughness=roughness_tensor, 
        sld=sld_tensor,
        abeles_func=abeles,
    )
    
    refl_curves_abeles_memory_eff = reflectivity(
        q=q, 
        thickness=thickness_tensor, 
        roughness=roughness_tensor, 
        sld=sld_tensor,
        abeles_func=abeles_memory_eff,
    )

    assert torch.allclose(refl_curves_abeles, refl_curves_abeles_memory_eff, atol=1e-10)


from tqdm import trange

from time import perf_counter

import torch
from torch import nn, Tensor

from reflectorch.data_generation.reflectivity import kinematical_approximation
from reflectorch.data_generation.priors.multilayer_structures import SimpleMultilayerSampler


class MultilayerFit(object):
    def __init__(self,
                 q: Tensor,
                 exp_curve: Tensor,
                 params: Tensor,
                 convert_func,
                 scale_curve_func,
                 min_bounds: Tensor,
                 max_bounds: Tensor,
                 optim_cls=None,
                 lr: float = 5e-2,
                 ):

        self.q = q
        self.scale_curve_func = scale_curve_func
        self.scaled_exp_curve = self.scale_curve_func(torch.atleast_2d(exp_curve))
        self.params_to_fit = nn.Parameter(params.clone())
        self.convert_func = convert_func

        self.min_bounds = min_bounds
        self.max_bounds = max_bounds

        optim_cls = optim_cls or torch.optim.Adam
        self.optim = optim_cls([self.params_to_fit], lr)

        self.best_loss = float('inf')
        self.best_params, self.best_curve = None, None

        self.get_best_solution()
        self.losses = []

    @classmethod
    def from_prior_sampler(
            cls,
            q: Tensor,
            exp_curve: Tensor,
            prior_sampler: SimpleMultilayerSampler,
            batch_size: int = 2 ** 13,
            **kwargs
    ):
        _, scaled_params = prior_sampler.optimized_sample(batch_size)
        params = prior_sampler.restore_params2parametrized(scaled_params)

        return cls(
            q.float(), exp_curve.float(), params.float(),
            convert_func=get_convert_func(prior_sampler.multilayer_model),
            scale_curve_func=_save_log, min_bounds=prior_sampler.min_bounds,
            max_bounds=prior_sampler.max_bounds,
            **kwargs
        )

    @classmethod
    def from_prediction(
            cls,
            param: Tensor,
            prior_sampler,
            q: Tensor,
            exp_curve: Tensor,
            batch_size: int = 2 ** 13,
            rel_bounds: float = 0.3,
            **kwargs
    ):

        num_params = param.shape[-1]

        deltas = prior_sampler.restore_params2parametrized(
            (torch.rand(batch_size, num_params, device=param.device) * 2 - 1) * rel_bounds
        ) - prior_sampler.min_bounds

        deltas[0] = 0.

        params = torch.atleast_2d(param) + deltas

        return cls(
            q.float(), exp_curve.float(), params.float(),
            convert_func=get_convert_func(prior_sampler.multilayer_model),
            scale_curve_func=_save_log, min_bounds=prior_sampler.min_bounds,
            max_bounds=prior_sampler.max_bounds,
            **kwargs
        )

    def get_clipped_params(self):
        return torch.clamp(self.params_to_fit.detach().clone(), self.min_bounds, self.max_bounds)

    def calc_loss(self, reduce=True, clip: bool = False):
        if clip:
            params = self.get_clipped_params()
        else:
            params = self.params_to_fit
        curves = self.get_curves(params=params)
        losses = ((self.scale_curve_func(curves) - self.scaled_exp_curve) ** 2)

        if reduce:
            return losses.sum()
        else:
            return losses.sum(-1)

    def get_curves(self, params: Tensor = None):
        if params is None:
            params = self.params_to_fit
        curves = kinematical_approximation(self.q, **self.convert_func(params))
        return torch.atleast_2d(curves)

    def run(self, num_iterations: int = 500, disable_tqdm: bool = False):
        pbar = trange(num_iterations, disable=disable_tqdm)

        for _ in pbar:
            self.optim.zero_grad()
            loss = self.calc_loss()
            loss.backward()
            self.optim.step()
            self.losses.append(loss.item())
            pbar.set_description(f'Loss = {loss.item():.2e}')

    def clear(self):
        self.losses.clear()

    def run_fixed_time(self, time_limit: float = 2.):

        start = perf_counter()

        while True:
            self.optim.zero_grad()
            loss = self.calc_loss()
            loss.backward()
            self.optim.step()
            self.losses.append(loss.item())
            time_spent = perf_counter() - start

            if time_spent > time_limit:
                break

    @torch.no_grad()
    def get_best_solution(self, clip: bool = True):
        losses = self.calc_loss(clip=clip, reduce=False)
        idx = torch.argmin(losses)
        best_loss = losses[idx].item()

        if best_loss < self.best_loss:
            best_curve = self.get_curves()[idx]
            best_params = self.params_to_fit.detach()[idx].clone()
            self.best_params, self.best_curve, self.best_loss = best_params, best_curve, best_loss

        return self.best_params


def get_convert_func(multilayer_model):
    def func(params):
        params = multilayer_model.to_standard_params(params)
        return {
            'thickness': params['thicknesses'],  # very useful transformation indeed ...
            'roughness': params['roughnesses'],
            'sld': params['slds'],
        }

    return func


def _save_log(curves, eps: float = 1e-10):
    return torch.log10(curves + eps)



from reflectorch.models.encoders import *
from reflectorch.models.networks import *

__all__ = [
    "ConvEncoder",
    "ConvDecoder",
    "ConvAutoencoder",
    "FnoEncoder",
    "IntegralConvEmbedding",
    "SpectralConv1d",
    "ConvResidualNet1D",
    "ResidualMLP",
    "NetworkWithPriors",
    "NetworkWithPriorsConvEmb",
    "NetworkWithPriorsFnoEmb",
]


from reflectorch.data_generation.dataset import BasicDataset, BATCH_DATA_TYPE
from reflectorch.data_generation.priors import (
    Params,
    PriorSampler,
    BasicPriorSampler,
    SingleParamPrior,
    SimplePriorSampler,
    UniformParamPrior,
    GaussianParamPrior,
    TruncatedGaussianParamPrior,
    UniformSubPriorParams,
    UniformSubPriorSampler,
    NarrowSldUniformSubPriorSampler,
    ExpUniformSubPriorSampler,
    SimpleMultilayerSampler,
    SubpriorParametricSampler,
    BasicParams,
    ParametricModel,
    MULTILAYER_MODELS,
    SamplerStrategy,
    BasicSamplerStrategy,
    ConstrainedRoughnessSamplerStrategy,
    ConstrainedRoughnessAndImgSldSamplerStrategy,
)
from reflectorch.data_generation.process_data import ProcessData, ProcessPipeline
from reflectorch.data_generation.q_generator import (
    QGenerator,
    ConstantAngle,
    ConstantQ,
    VariableQ,
    EquidistantQ,
    MaskedVariableQ,
)
from reflectorch.data_generation.noise import (
    QNoiseGenerator,
    IntensityNoiseGenerator,
    QNormalNoiseGenerator,
    QSystematicShiftGenerator,
    PoissonNoiseGenerator,
    MultiplicativeLogNormalNoiseGenerator,
    ShiftNoise,
    ScalingNoise,
    BackgroundNoise,
    BasicExpIntensityNoise,
    GaussianExpIntensityNoise,
    BasicQNoiseGenerator,
)
from reflectorch.data_generation.scale_curves import (
    CurvesScaler,
    LogAffineCurvesScaler,
    MeanNormalizationCurvesScaler,
)
from reflectorch.data_generation.utils import (
    get_reversed_params,
    get_density_profiles,
    uniform_sampler,
    logdist_sampler,
    triangular_sampler,
    get_param_labels,
)

from reflectorch.data_generation.smearing import Smearing

from reflectorch.data_generation.reflectivity import reflectivity

from reflectorch.data_generation.likelihoods import (
    LogLikelihood,
    PoissonLogLikelihood,
)

__all__ = [
    "Params",
    "PriorSampler",
    "BasicPriorSampler",
    "BasicDataset",
    "ProcessData",
    "ProcessPipeline",
    "QGenerator",
    "ConstantQ",
    "VariableQ",
    "EquidistantQ",
    "MaskedVariableQ",
    "QNoiseGenerator",
    "IntensityNoiseGenerator",
    "MultiplicativeLogNormalNoiseGenerator",
    "PoissonNoiseGenerator",
    "CurvesScaler",
    "ShiftNoise",
    "ScalingNoise",
    "BackgroundNoise",
    "QNormalNoiseGenerator",
    "QSystematicShiftGenerator",
    "LogAffineCurvesScaler",
    "MeanNormalizationCurvesScaler",
    "get_reversed_params",
    "get_density_profiles",
    "logdist_sampler",
    "uniform_sampler",
    "triangular_sampler",
    "get_param_labels",
    "reflectivity",
    "Smearing",
    "SingleParamPrior",
    "SimplePriorSampler",
    "UniformParamPrior",
    "GaussianParamPrior",
    "TruncatedGaussianParamPrior",
    "UniformSubPriorParams",
    "UniformSubPriorSampler",
    "NarrowSldUniformSubPriorSampler",
    "ExpUniformSubPriorSampler",
    "SimpleMultilayerSampler",
    "BATCH_DATA_TYPE",
    "LogLikelihood",
    "PoissonLogLikelihood",
    "BasicExpIntensityNoise",
    "GaussianExpIntensityNoise",
    "BasicQNoiseGenerator",
    "ConstantAngle",
    "SubpriorParametricSampler",
    "BasicParams",
    "ParametricModel",
    "MULTILAYER_MODELS",
    "SamplerStrategy",
    "BasicSamplerStrategy",
    "ConstrainedRoughnessSamplerStrategy",
    "ConstrainedRoughnessAndImgSldSamplerStrategy",
]



from pathlib import Path
import numpy as np

def load_mft_data(filepath):
    """
    Load q, reflectivity, reflectivity error, and q-resolution from an .mft file.
    
    Parameters:
        filepath (str or Path): Path to the .mft file

    Returns:
        q, refl, refl_err, q_res : np.ndarray
    """
    filepath = Path(filepath)
    
    with filepath.open('r', encoding='utf-8') as f:
        lines = f.readlines()

    start_idx = next(
        i for i, line in enumerate(lines)
        if line.strip().startswith('q') and 'q_res' in line
    ) + 1

    data = []
    for line in lines[start_idx:]:
        parts = line.strip().split()
        if len(parts) == 4:
            try:
                data.append([float(p.replace('E', 'e')) for p in parts])
            except ValueError:
                continue

    if not data:
        raise ValueError(f"No valid data found in {filepath}")

    data_array = np.array(data)
    return data_array.T


try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np
from numpy import ndarray

NORMALIZE_MODE = Literal["first", "max", "incoming_intensity"]


def intensity2reflectivity(intensity: ndarray, mode: NORMALIZE_MODE, incoming_intensity=None) -> np.ndarray:
    if mode == "first":
        return intensity / intensity[0]
    if mode == "max":
        return intensity / intensity.max()
    if mode == "incoming_intensity":
        if incoming_intensity is None:
            raise ValueError("incoming_intensity is None")
        return intensity / incoming_intensity
    raise ValueError(f"Unknown mode {mode}")



import torch
import torch.nn as nn
import torch.nn.functional as F

from reflectorch.models.activations import activation_by_name

class SpectralConv1d(nn.Module):
    def __init__(self, in_channels, out_channels, modes):
        super().__init__()

        """
        1D Fourier layer. It does FFT, linear transform, and Inverse FFT.    
        """

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.modes = modes  #Number of Fourier modes to multiply, at most floor(N/2) + 1

        self.scale = (1 / (in_channels*out_channels))
        self.weights1 = nn.Parameter(self.scale * torch.rand(in_channels, out_channels, modes, dtype=torch.cfloat))

    # Complex multiplication
    def compl_mul1d(self, input, weights):
        # (batch, in_channel, x ), (in_channel, out_channel, x) -> (batch, out_channel, x)
        return torch.einsum("bix,iox->box", input, weights)

    def forward(self, x):
        batchsize = x.shape[0]
        #Compute Fourier coeffcients up to factor of e^(- something constant)
        x_ft = torch.fft.rfft(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batchsize, self.out_channels, x.size(-1)//2 + 1,  device=x.device, dtype=torch.cfloat)
        out_ft[:, :, :self.modes] = self.compl_mul1d(x_ft[:, :, :self.modes], self.weights1)

        #Return to physical space
        x = torch.fft.irfft(out_ft, n=x.size(-1))
        return x
    

class FnoEncoder(nn.Module):
    """An embedding network based on the Fourier Neural Operator (FNO) architecture

    .. image:: ../documentation/fig_reflectometry_embedding_networks.png
        :width: 400px
        :align: center

    Args:
        in_channels (int): number of input channels
        dim_embedding (int): dimension of the output embedding
        modes (int): number of Fourier modes
        width_fno (int): number of channels of the intermediate representations
        n_fno_blocks (int): number of FNO blocks
        activation (str): the activation function
        fusion_self_attention (bool): whether to use fusion self attention for merging the tokens (instead of mean)
        fsa_activation (str): the activation function of the fusion self attention block
    """
    def __init__(
            self, 
            in_channels: int = 2, 
            dim_embedding: int = 128, 
            modes: int = 32, 
            width_fno: int = 64, 
            n_fno_blocks: int = 6, 
            activation: str = 'gelu',
            fusion_self_attention: bool = False,
            fsa_activation: str = 'tanh',
            ):
        super().__init__()


        self.in_channels = in_channels
        self.dim_embedding = dim_embedding
        
        self.modes = modes
        self.width_fno = width_fno
        self.n_fno_blocks = n_fno_blocks
        self.activation = activation_by_name(activation)()
        self.fusion_self_attention = fusion_self_attention
        

        self.fc0 = nn.Linear(in_channels, width_fno) #(r(q), q)
        self.spectral_convs = nn.ModuleList([
            SpectralConv1d(in_channels=width_fno, out_channels=width_fno, modes=modes) for _ in range(n_fno_blocks)
            ])
        self.w_convs = nn.ModuleList([
            nn.Conv1d(in_channels=width_fno, out_channels=width_fno, kernel_size=1) for _ in range(n_fno_blocks)
            ])
        self.fc_out = nn.Linear(width_fno, dim_embedding)

        if fusion_self_attention:
            self.fusion = FusionSelfAttention(embed_dim=width_fno, hidden_dim=2*width_fno, activation=fsa_activation)
        
    def forward(self, x):
        """"""

        x = x.permute(0, 2, 1) #(B, D, S) -> (B, S, D)
        x = self.fc0(x)
        x = x.permute(0, 2, 1) #(B, S, D) -> (B, D, S) 

        for i in range(self.n_fno_blocks):
            x1 = self.spectral_convs[i](x)
            x2 = self.w_convs[i](x)
            
            x = x1 + x2
            x = self.activation(x)

        if self.fusion_self_attention:
            x = x.permute(0, 2, 1)
            x = self.fusion(x)  
        else:
            x = x.mean(dim=-1)

        x = self.fc_out(x)
        
        return x
    

class FusionSelfAttention(nn.Module):
    def __init__(self, embed_dim: int = 64, hidden_dim: int = 64, activation: str = 'gelu'):
        super().__init__()
        activation = activation_by_name(activation)()
        self.fuser = nn.Sequential(nn.Linear(embed_dim, hidden_dim), 
                                   activation,
                                   nn.Linear(hidden_dim, 1, bias=False))
        
    def forward(self, 
                c: torch.Tensor,  # (batch_size x seq_len x embed_dim)
                mask: torch.Tensor = None, # (batch_size x seq_len)
                ):  
        a = self.fuser(c)
        alpha = torch.exp(a)*mask.unsqueeze(-1) if mask is not None else torch.exp(a)
        alpha = alpha/alpha.sum(dim=1, keepdim=True)
        return (alpha*c).sum(dim=1)  # (batch_size x embed_dim)


from typing import Union
from pathlib import Path

__all__ = [
    'ROOT_DIR',
    'SAVED_MODELS_DIR',
    'SAVED_LOSSES_DIR',
    'RUN_SCRIPTS_DIR',
    'CONFIG_DIR',
    'TESTS_PATH',
    'TEST_DATA_PATH',
    'listdir',
]

ROOT_DIR: Path = Path(__file__).parents[1]
SAVED_MODELS_DIR: Path = ROOT_DIR / 'saved_models'
SAVED_LOSSES_DIR: Path = ROOT_DIR / 'saved_losses'
RUN_SCRIPTS_DIR: Path = ROOT_DIR / 'runs'
CONFIG_DIR: Path = ROOT_DIR / 'configs'
TESTS_PATH: Path = ROOT_DIR / 'tests'
TEST_DATA_PATH = TESTS_PATH / 'data'


def listdir(path: Union[Path, str], pattern: str = '*', recursive: bool = False, *, sort_key=None, reverse=False):
    path = Path(path)
    func = path.rglob if recursive else path.glob
    return sorted(list(func(pattern)), key=sort_key, reverse=reverse)



from typing import Tuple, Dict, List

import torch
from torch import Tensor

from reflectorch.data_generation.reflectivity import (
    reflectivity,
    abeles_memory_eff,
    kinematical_approximation,
)
from reflectorch.data_generation.utils import (
    get_param_labels,
)
from reflectorch.data_generation.priors.sampler_strategies import (
    SamplerStrategy,
    BasicSamplerStrategy,
    ConstrainedRoughnessSamplerStrategy,
    ConstrainedRoughnessAndImgSldSamplerStrategy,
)

__all__ = [
    "MULTILAYER_MODELS",
    "ParametricModel",
]


class ParametricModel(object):
    """Base class for parameterizations of the SLD profile.

    Args:
        max_num_layers (int): the number of layers
    """
    NAME: str = ''
    PARAMETER_NAMES: Tuple[str, ...]

    def __init__(self, max_num_layers: int, **kwargs):
        self.max_num_layers = max_num_layers
        self._sampler_strategy = self._init_sampler_strategy(**kwargs)

    def _init_sampler_strategy(self, nuisance_params_dim: int = 0, **kwargs):
        return BasicSamplerStrategy(**kwargs)

    @property
    def param_dim(self) -> int:
        """get the number of parameters

        Returns:
            int:
        """
        return len(self.PARAMETER_NAMES)

    @property
    def sampler_strategy(self) -> SamplerStrategy:
        """get the sampler strategy

        Returns:
            SamplerStrategy:
        """
        return self._sampler_strategy

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        """computes the reflectivity curves 

        Args:
            q: the reciprocal space (q) positions
            parametrized_model (Tensor): the values of the parameters

        Returns:
            Tensor: the computed reflectivity curves
        """
        params = self.to_standard_params(parametrized_model)
        return reflectivity(q, **params, **kwargs)

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        raise NotImplementedError

    def from_standard_params(self, params: dict) -> Tensor:
        raise NotImplementedError

    def scale_with_q(self, parametrized_model: Tensor, q_ratio: float) -> Tensor:
        raise NotImplementedError

    def init_bounds(self,
                    param_ranges: Dict[str, Tuple[float, float]],
                    bound_width_ranges: Dict[str, Tuple[float, float]],
                    device=None,
                    dtype=None,
                    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """initializes arrays storing individually the upper and lower bounds from the dictionaries of parameter and bound width ranges 

        Args:
            param_ranges (Dict[str, Tuple[float, float]]): parameter ranges
            bound_width_ranges (Dict[str, Tuple[float, float]]): bound width ranges
            device (optional): the Pytorch device. Defaults to None.
            dtype (optional): the Pytorch datatype. Defaults to None.

        Returns:
            Tuple[Tensor, Tensor, Tensor, Tensor]:
        """
        ordered_bounds = [param_ranges[k] for k in self.PARAMETER_NAMES]
        delta_bounds = [bound_width_ranges[k] for k in self.PARAMETER_NAMES]

        min_bounds, max_bounds = torch.tensor(ordered_bounds, device=device, dtype=dtype).T[:, None]
        min_deltas, max_deltas = torch.tensor(delta_bounds, device=device, dtype=dtype).T[:, None]

        return min_bounds, max_bounds, min_deltas, max_deltas

    def get_param_labels(self, **kwargs) -> List[str]:
        """get the list with the name of the parameters

        Returns:
            List[str]:
        """
        return list(self.PARAMETER_NAMES)

    def sample(self, batch_size: int,
               total_min_bounds: Tensor,
               total_max_bounds: Tensor,
               total_min_delta: Tensor,
               total_max_delta: Tensor,
               ):
        """samples the parameter values and their prior bounds

        Args:
            batch_size (int): the batch size
            total_min_bounds (Tensor): lower bounds of the parameter ranges
            total_max_bounds (Tensor): upper bounds of the parameter ranges
            total_min_delta (Tensor): lower widths of the subprior intervals
            total_max_delta (Tensor): upper widths of the subprior intervals

        Returns:
            Tensor: sampled parameters
        """
        return self.sampler_strategy.sample(
            batch_size,
            total_min_bounds,
            total_max_bounds,
            total_min_delta,
            total_max_delta,
        )


class StandardModel(ParametricModel):
    """Parameterization for the standard box model. The parameters are the thicknesses, roughnesses and real sld values of the layers."""
    NAME = 'standard_model'

    PARAMETER_NAMES = (
        "thicknesses",
        "roughnesses",
        "slds",
    )

    @property
    def param_dim(self) -> int:
        return 3 * self.max_num_layers + 2

    def _init_sampler_strategy(self,
                               constrained_roughness: bool = True,
                               max_thickness_share: float = 0.5,
                               nuisance_params_dim: int = 0,
                               **kwargs):
        if constrained_roughness:
            num_params = self.param_dim + nuisance_params_dim
            thickness_mask = torch.zeros(num_params, dtype=torch.bool)
            roughness_mask = torch.zeros(num_params, dtype=torch.bool)
            thickness_mask[:self.max_num_layers] = True
            roughness_mask[self.max_num_layers:2 * self.max_num_layers + 1] = True
            return ConstrainedRoughnessSamplerStrategy(
                thickness_mask, roughness_mask,
                max_thickness_share=max_thickness_share,
                **kwargs
            )
        else:
            return BasicSamplerStrategy(**kwargs)

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        return self._params2dict(parametrized_model)

    def init_bounds(self,
                    param_ranges: Dict[str, Tuple[float, float]],
                    bound_width_ranges: Dict[str, Tuple[float, float]],
                    device=None,
                    dtype=None,
                    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:

        other_ranges = [param_ranges[k] for k in self.PARAMETER_NAMES[3:]]
        other_delta_bounds = [bound_width_ranges[k] for k in self.PARAMETER_NAMES[3:]]

        ordered_bounds = (
                [param_ranges["thicknesses"]] * self.max_num_layers +
                [param_ranges["roughnesses"]] * (self.max_num_layers + 1) +
                [param_ranges["slds"]] * (self.max_num_layers + 1) +
                other_ranges
        )
        delta_bounds = (
                [bound_width_ranges["thicknesses"]] * self.max_num_layers +
                [bound_width_ranges["roughnesses"]] * (self.max_num_layers + 1) +
                [bound_width_ranges["slds"]] * (self.max_num_layers + 1) +
                other_delta_bounds
        )

        min_bounds, max_bounds = torch.tensor(ordered_bounds, device=device, dtype=dtype).T[:, None]
        min_deltas, max_deltas = torch.tensor(delta_bounds, device=device, dtype=dtype).T[:, None]

        return min_bounds, max_bounds, min_deltas, max_deltas

    def get_param_labels(self, **kwargs) -> List[str]:
        return get_param_labels(self.max_num_layers, **kwargs)

    @staticmethod
    def _params2dict(parametrized_model: Tensor):
        num_params = parametrized_model.shape[-1]
        num_layers = (num_params - 2) // 3
        assert num_layers * 3 + 2 == num_params

        d, sigma, sld = torch.split(
            parametrized_model, [num_layers, num_layers + 1, num_layers + 1], -1
        )
        params = dict(
            thickness=d,
            roughness=sigma,
            sld=sld
        )

        return params

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        return reflectivity(
            q, **self._params2dict(parametrized_model), **kwargs
        )


class ModelWithAbsorption(StandardModel):
    """Parameterization for the box model in which the imaginary sld values of the layers are additional parameters."""
    NAME = 'model_with_absorption'

    PARAMETER_NAMES = (
        "thicknesses",
        "roughnesses",
        "slds",
        "islds",
    )

    @property
    def param_dim(self) -> int:
        return 4 * self.max_num_layers + 3
    
    def _init_sampler_strategy(self,
                               constrained_roughness: bool = True,
                               constrained_isld: bool = True,
                               max_thickness_share: float = 0.5,
                               max_sld_share: float = 0.2,
                               nuisance_params_dim: int = 0,
                               **kwargs):
        if constrained_roughness:
            num_params = self.param_dim + nuisance_params_dim
            thickness_mask = torch.zeros(num_params, dtype=torch.bool)
            roughness_mask = torch.zeros(num_params, dtype=torch.bool)
            thickness_mask[:self.max_num_layers] = True
            roughness_mask[self.max_num_layers:2 * self.max_num_layers + 1] = True

            if constrained_isld:
                sld_mask = torch.zeros(num_params, dtype=torch.bool)
                isld_mask = torch.zeros(num_params, dtype=torch.bool)
                sld_mask[2 * self.max_num_layers + 1:3 * self.max_num_layers + 2] = True
                isld_mask[3 * self.max_num_layers + 2:4 * self.max_num_layers + 3] = True
                return ConstrainedRoughnessAndImgSldSamplerStrategy(
                    thickness_mask, roughness_mask, sld_mask, isld_mask,
                    max_thickness_share=max_thickness_share, max_sld_share=max_sld_share,
                    **kwargs
                )
            else:
                return ConstrainedRoughnessSamplerStrategy(
                    thickness_mask, roughness_mask,
                    max_thickness_share=max_thickness_share,
                    **kwargs
            )
        else:
            return BasicSamplerStrategy(**kwargs)

    def init_bounds(self,
                    param_ranges: Dict[str, Tuple[float, float]],
                    bound_width_ranges: Dict[str, Tuple[float, float]],
                    device=None,
                    dtype=None,
                    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        other_ranges = [param_ranges[k] for k in self.PARAMETER_NAMES[4:]]
        other_delta_bounds = [bound_width_ranges[k] for k in self.PARAMETER_NAMES[4:]]

        ordered_bounds = (
                [param_ranges["thicknesses"]] * self.max_num_layers +
                [param_ranges["roughnesses"]] * (self.max_num_layers + 1) +
                [param_ranges["slds"]] * (self.max_num_layers + 1) +
                [param_ranges["islds"]] * (self.max_num_layers + 1) +
                other_ranges
        )
        delta_bounds = (
                [bound_width_ranges["thicknesses"]] * self.max_num_layers +
                [bound_width_ranges["roughnesses"]] * (self.max_num_layers + 1) +
                [bound_width_ranges["slds"]] * (self.max_num_layers + 1) +
                [bound_width_ranges["islds"]] * (self.max_num_layers + 1) +
                other_delta_bounds
        )

        min_bounds, max_bounds = torch.tensor(ordered_bounds, device=device, dtype=dtype).T[:, None]
        min_deltas, max_deltas = torch.tensor(delta_bounds, device=device, dtype=dtype).T[:, None]

        return min_bounds, max_bounds, min_deltas, max_deltas

    def get_param_labels(self, **kwargs) -> List[str]:
        return get_param_labels(self.max_num_layers, parameterization_type='absorption', **kwargs)
    
    @staticmethod
    def _params2dict(parametrized_model: Tensor):
        num_params = parametrized_model.shape[-1]
        num_layers = (num_params - 3) // 4
        assert num_layers * 4 + 3 == num_params

        d, sigma, sld, isld = torch.split(
            parametrized_model, [num_layers, num_layers + 1, num_layers + 1, num_layers + 1], -1
        )
        params = dict(
            thickness=d,
            roughness=sigma,
            sld=sld + 1j * isld
        )

        return params

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        return reflectivity(
            q, **self._params2dict(parametrized_model), **kwargs
        )


class ModelWithShifts(StandardModel):
    """Variant of the standard box model parameterization in which two additional parameters are considered: the shift in the q positions (additive) and the shift in 
    intensity (multiplicative, or additive in log domain)."""
    NAME = 'model_with_shifts'

    PARAMETER_NAMES = (
        "thicknesses",
        "roughnesses",
        "slds",
        "q_shift",
        "norm_shift",
    )

    @property
    def param_dim(self) -> int:
        return 3 * self.max_num_layers + 4

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        params = self._params2dict(parametrized_model)
        params.pop('q_shift')
        params.pop('norm_shift')

        return params

    def get_param_labels(self, **kwargs) -> List[str]:
        return get_param_labels(self.max_num_layers, **kwargs) + [r"$\Delta q$ (Å$^{{-1}}$)", r"$\Delta I$"]


    @staticmethod
    def _params2dict(parametrized_model: Tensor):
        num_params = parametrized_model.shape[-1]
        num_layers = (num_params - 4) // 3
        assert num_layers * 3 + 4 == num_params

        d, sigma, sld, q_shift, norm_shift = torch.split(
            parametrized_model, [num_layers, num_layers + 1, num_layers + 1, 1, 1], -1
        )
        params = dict(
            thickness=d,
            roughness=sigma,
            sld=sld,
            q_shift=q_shift,
            norm_shift=norm_shift,
        )

        return params

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        return reflectivity_with_shifts(
            q, **self._params2dict(parametrized_model), **kwargs
        )

def reflectivity_with_shifts(q, thickness, roughness, sld, q_shift, norm_shift, **kwargs):
    q = torch.atleast_2d(q) + q_shift
    return reflectivity(q, thickness, roughness, sld, **kwargs) * norm_shift   

class NoFresnelModel(StandardModel):
    NAME = 'no_fresnel_model'

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        return kinematical_approximation(
            q, **self._params2dict(parametrized_model), apply_fresnel=False, **kwargs
        )


class BasicMultilayerModel1(ParametricModel):
    NAME = 'repeating_multilayer_v1'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
    )

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        return multilayer_model1(parametrized_model, self.max_num_layers)

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        params = self.to_standard_params(parametrized_model)
        return reflectivity(q, abeles_func=abeles_memory_eff, **params, **kwargs)


class BasicMultilayerModel2(BasicMultilayerModel1):
    NAME = 'repeating_multilayer_v2'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "dr_sigmoid_rel_pos",
        "dr_sigmoid_rel_width",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
    )

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        return multilayer_model2(parametrized_model, self.max_num_layers)


class BasicMultilayerModel3(BasicMultilayerModel1):
    """Parameterization for a thin film composed of repeating identical monolayers, each monolayer consisting of two boxes with distinct SLDs. 
    A sigmoid envelope modulating the SLD profile of the monolayers defines the film thickness and the roughness at the top interface. 
    A second sigmoid envelope can be used to modulate the amplitude of the monolayer SLDs as a function of the displacement from the position of the first sigmoid. 
    These two sigmoids allow one to model a thin film that is coherently ordered up to a certain coherent thickness and gets incoherently ordered or amorphous toward the top of the film.
    In addition, a layer between the substrate and the multilayer (”phase layer”) is introduced to account for the interface structure, 
    which does not necessarily have to be identical to the multilayer period.
    """

    NAME = 'repeating_multilayer_v3'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "dr_sigmoid_rel_pos",
        "dr_sigmoid_rel_width",
        "d_block1_rel",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
    )

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        return multilayer_model3(parametrized_model, self.max_num_layers)


class MultilayerModel1WithShifts(BasicMultilayerModel1):
    NAME = 'repeating_multilayer_v1_with_shifts'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
        "q_shift",
        "norm_shift",
    )

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        q_shift, norm_shift = parametrized_model[..., -2:].T[..., None]
        return reflectivity_with_shifts(
            q, q_shift=q_shift, norm_shift=norm_shift, abeles_func=abeles_memory_eff,
            **self.to_standard_params(parametrized_model), **kwargs
        )


class MultilayerModel3WithShifts(BasicMultilayerModel3):
    NAME = 'repeating_multilayer_v3_with_shifts'

    PARAMETER_NAMES = (
        "d_full_rel",
        "rel_sigmas",
        "dr_sigmoid_rel_pos",
        "dr_sigmoid_rel_width",
        "d_block1_rel",
        "d_block",
        "s_block_rel",
        "r_block",
        "dr",
        "d3_rel",
        "s3_rel",
        "r3",
        "d_sio2",
        "s_sio2",
        "s_si",
        "r_sio2",
        "r_si",
        "q_shift",
        "norm_shift",
    )

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        q_shift, norm_shift = parametrized_model[..., -2:].T[..., None]
        return reflectivity_with_shifts(
            q, q_shift=q_shift, norm_shift=norm_shift, abeles_func=abeles_memory_eff,
            **self.to_standard_params(parametrized_model), **kwargs
        )


MULTILAYER_MODELS = {
    'standard_model': StandardModel,
    'model_with_absorption': ModelWithAbsorption,
    'model_with_shifts': ModelWithShifts,
    'no_fresnel_model': NoFresnelModel,
    'repeating_multilayer_v1': BasicMultilayerModel1,
    'repeating_multilayer_v2': BasicMultilayerModel2,
    'repeating_multilayer_v3': BasicMultilayerModel3,
    'repeating_multilayer_v1_with_shifts': MultilayerModel1WithShifts,
    'repeating_multilayer_v3_with_shifts': MultilayerModel3WithShifts,
}


def multilayer_model1(parametrized_model: Tensor, d_full_rel_max: int = 30) -> dict:
    n = d_full_rel_max

    (
        d_full_rel,
        rel_sigmas,
        d_block,
        s_block_rel,
        r_block,
        dr,
        d3_rel,
        s3_rel,
        r3,
        d_sio2,
        s_sio2,
        s_si,
        r_sio2,
        r_si,
        *_,
    ) = parametrized_model.T

    batch_size = parametrized_model.shape[0]

    r_positions = 2 * n - torch.arange(2 * n, dtype=dr.dtype, device=dr.device)[None].repeat(batch_size, 1)

    r_modulations = torch.sigmoid(-(r_positions - 2 * d_full_rel[..., None]) / rel_sigmas[..., None])

    r_block = r_block[:, None].repeat(1, n)
    dr = dr[:, None].repeat(1, n)

    sld_blocks = torch.stack([r_block, r_block + dr], -1).flatten(1)

    sld_blocks = r_modulations * sld_blocks

    d3 = d3_rel * d_block

    thicknesses = torch.cat(
        [(d_block / 2)[:, None].repeat(1, n * 2), d3[:, None], d_sio2[:, None]], -1
    )

    s_block = s_block_rel * d_block

    roughnesses = torch.cat(
        [s_block[:, None].repeat(1, n * 2), (s3_rel * d3)[:, None], s_sio2[:, None], s_si[:, None]], -1
    )

    slds = torch.cat(
        [sld_blocks, r3[:, None], r_sio2[:, None], r_si[:, None]], -1
    )

    params = dict(
        thickness=thicknesses,
        roughness=roughnesses,
        sld=slds
    )
    return params


def multilayer_model2(parametrized_model: Tensor, d_full_rel_max: int = 30) -> dict:
    n = d_full_rel_max

    (
        d_full_rel,
        rel_sigmas,
        dr_sigmoid_rel_pos,
        dr_sigmoid_rel_width,
        d_block,
        s_block_rel,
        r_block,
        dr,
        d3_rel,
        s3_rel,
        r3,
        d_sio2,
        s_sio2,
        s_si,
        r_sio2,
        r_si,
        *_,
    ) = parametrized_model.T

    batch_size = parametrized_model.shape[0]

    r_positions = 2 * n - torch.arange(2 * n, dtype=dr.dtype, device=dr.device)[None].repeat(batch_size, 1)

    r_modulations = torch.sigmoid(-(r_positions - 2 * d_full_rel[..., None]) / rel_sigmas[..., None])

    r_block = r_block[:, None].repeat(1, n)
    dr = dr[:, None].repeat(1, n)

    dr_positions = r_positions[:, ::2]

    dr_modulations = torch.sigmoid(
        -(dr_positions - (2 * d_full_rel * dr_sigmoid_rel_pos)[..., None]) / dr_sigmoid_rel_width[..., None]
    )

    dr = dr * dr_modulations

    sld_blocks = torch.stack([r_block, r_block + dr], -1).flatten(1)

    sld_blocks = r_modulations * sld_blocks

    d3 = d3_rel * d_block

    thicknesses = torch.cat(
        [(d_block / 2)[:, None].repeat(1, n * 2), d3[:, None], d_sio2[:, None]], -1
    )

    s_block = s_block_rel * d_block

    roughnesses = torch.cat(
        [s_block[:, None].repeat(1, n * 2), (s3_rel * d3)[:, None], s_sio2[:, None], s_si[:, None]], -1
    )

    slds = torch.cat(
        [sld_blocks, r3[:, None], r_sio2[:, None], r_si[:, None]], -1
    )

    params = dict(
        thickness=thicknesses,
        roughness=roughnesses,
        sld=slds
    )
    return params


def multilayer_model3(parametrized_model: Tensor, d_full_rel_max: int = 30):
    n = d_full_rel_max

    (
        d_full_rel,
        rel_sigmas,
        dr_sigmoid_rel_pos,
        dr_sigmoid_rel_width,
        d_block1_rel,
        d_block,
        s_block_rel,
        r_block,
        dr,
        d3_rel,
        s3_rel,
        r3,
        d_sio2,
        s_sio2,
        s_si,
        r_sio2,
        r_si,
        *_,
    ) = parametrized_model.T

    batch_size = parametrized_model.shape[0]

    r_positions = 2 * n - torch.arange(2 * n, dtype=dr.dtype, device=dr.device)[None].repeat(batch_size, 1)

    r_modulations = torch.sigmoid(
        -(
                r_positions - 2 * d_full_rel[..., None]
        ) / rel_sigmas[..., None]
    )

    dr_positions = r_positions[:, ::2]

    dr_modulations = dr[..., None] * (1 - torch.sigmoid(
        -(
                dr_positions - 2 * d_full_rel[..., None] + 2 * dr_sigmoid_rel_pos[..., None]
        ) / dr_sigmoid_rel_width[..., None]
    ))

    r_block = r_block[..., None].repeat(1, n)
    dr = dr[..., None].repeat(1, n)

    sld_blocks = torch.stack(
        [
            r_block + dr_modulations * (1 - d_block1_rel[..., None]),
            r_block + dr - dr_modulations * d_block1_rel[..., None]
        ], -1).flatten(1)

    sld_blocks = r_modulations * sld_blocks

    d3 = d3_rel * d_block

    d1, d2 = d_block * d_block1_rel, d_block * (1 - d_block1_rel)

    thickness_blocks = torch.stack([d1[:, None].repeat(1, n), d2[:, None].repeat(1, n)], -1).flatten(1)

    thicknesses = torch.cat(
        [thickness_blocks, d3[:, None], d_sio2[:, None]], -1
    )

    s_block = s_block_rel * d_block

    roughnesses = torch.cat(
        [s_block[:, None].repeat(1, n * 2), (s3_rel * d3)[:, None], s_sio2[:, None], s_si[:, None]], -1
    )

    slds = torch.cat(
        [sld_blocks, r3[:, None], r_sio2[:, None], r_si[:, None]], -1
    )

    params = dict(
        thickness=thicknesses,
        roughness=roughnesses,
        sld=slds
    )
    return params


class NuisanceParamsWrapper(ParametricModel):
    """
    Wraps a base model (e.g. StandardModel) to add nuisance parameters, allowing independent enabling/disabling.

    Args:
        base_model (ParametricModel): The base parametric model.
        nuisance_params_config (Dict[str, bool]): Dictionary where keys are parameter names
                                                  and values are `True` (enable) or `False` (disable).
    """

    def __init__(self, base_model: ParametricModel, nuisance_params_config: Dict[str, bool] = None, **kwargs):
        self.base_model = base_model
        self.nuisance_params_config = nuisance_params_config or {}

        self.enabled_nuisance_params = [name for name, is_enabled in self.nuisance_params_config.items() if is_enabled]

        self.PARAMETER_NAMES = self.base_model.PARAMETER_NAMES + tuple(self.enabled_nuisance_params)
        self._param_dim = self.base_model.param_dim + len(self.enabled_nuisance_params)
        
        super().__init__(base_model.max_num_layers, **kwargs)

    def _init_sampler_strategy(self, **kwargs):
        return self.base_model._init_sampler_strategy(nuisance_params_dim=len(self.enabled_nuisance_params), **kwargs)

    @property
    def param_dim(self) -> int:
        return self._param_dim

    def to_standard_params(self, parametrized_model: Tensor) -> dict:
        """Extracts base model parameters only."""
        base_dim = self.base_model.param_dim
        base_part = parametrized_model[..., :base_dim]
        return self.base_model.to_standard_params(base_part)

    def reflectivity(self, q, parametrized_model: Tensor, **kwargs) -> Tensor:
        """Computes reflectivity with optional nuisance parameter shifts."""
        base_dim = self.base_model.param_dim
        base_params = parametrized_model[..., :base_dim]
        nuisance_part = parametrized_model[..., base_dim:]

        nuisance_dict = {param: nuisance_part[..., i].unsqueeze(-1) for i, param in enumerate(self.enabled_nuisance_params)}
        if "log10_background" in nuisance_dict:
            nuisance_dict["background"] = 10 ** nuisance_dict.pop("log10_background")

        return self.base_model.reflectivity(q, base_params, **nuisance_dict, **kwargs)

    def init_bounds(self, param_ranges: Dict[str, Tuple[float, float]],
                    bound_width_ranges: Dict[str, Tuple[float, float]], device=None, dtype=None):
        """Initialize bounds for enabled nuisance parameters."""
        min_bounds_base, max_bounds_base, min_deltas_base, max_deltas_base = self.base_model.init_bounds(
            param_ranges, bound_width_ranges, device, dtype)

        ordered_bounds_nuisance = [param_ranges[k] for k in self.enabled_nuisance_params]
        delta_bounds_nuisance = [bound_width_ranges[k] for k in self.enabled_nuisance_params]

        if ordered_bounds_nuisance:
            min_bounds_nuisance, max_bounds_nuisance = torch.tensor(ordered_bounds_nuisance, device=device, dtype=dtype).T[:, None]
            min_deltas_nuisance, max_deltas_nuisance = torch.tensor(delta_bounds_nuisance, device=device, dtype=dtype).T[:, None]

            min_bounds = torch.cat([min_bounds_base, min_bounds_nuisance], dim=-1)
            max_bounds = torch.cat([max_bounds_base, max_bounds_nuisance], dim=-1)
            min_deltas = torch.cat([min_deltas_base, min_deltas_nuisance], dim=-1)
            max_deltas = torch.cat([max_deltas_base, max_deltas_nuisance], dim=-1)
        else:
            min_bounds, max_bounds, min_deltas, max_deltas = min_bounds_base, max_bounds_base, min_deltas_base, max_deltas_base

        return min_bounds, max_bounds, min_deltas, max_deltas
    
    def get_param_labels(self, **kwargs) -> List[str]:
        return self.base_model.get_param_labels(**kwargs) + self.enabled_nuisance_params